"""
Evaluation pipeline classes for generated molecules.
"""

import sys
import os
from typing import Tuple, Optional
from pathlib import Path
from copy import deepcopy
from importlib.metadata import distributions

import numpy as np
import pandas as pd
from rdkit import Chem

if any(d.metadata["Name"] == 'rdkit' for d in distributions()):
    from rdkit.Contrib.SA_Score import sascorer # type: ignore
else:
    sys.path.append(os.path.join(os.environ['CONDA_PREFIX'],'share','RDKit','Contrib'))
    from SA_Score import sascorer # type: ignore

from rdkit.Chem import QED, Crippen, Lipinski, rdFingerprintGenerator
from rdkit.Chem.rdMolAlign import GetBestRMS, AlignMol

from shepherd_score.evaluations.utils.convert_data import extract_mol_from_xyz_block, get_mol_from_atom_pos

from shepherd_score.score.constants import ALPHA, LAM_SCALING
from shepherd_score.score.constants import P_TYPES

from shepherd_score.conformer_generation import optimize_conformer_with_xtb_from_xyz_block, single_point_xtb_from_xyz

from shepherd_score.container import Molecule, MoleculePair
from shepherd_score.score.gaussian_overlap_np import get_overlap_np
from shepherd_score.score.electrostatic_scoring_np import get_overlap_esp_np
from shepherd_score.score.pharmacophore_scoring_np import get_overlap_pharm_np

RNG = np.random.default_rng()
morgan_fp_gen = rdFingerprintGenerator.GetMorganGenerator(radius=3, includeChirality=True)

TMPDIR = Path('./')
if 'TMPDIR' in os.environ:
    TMPDIR = Path(os.environ['TMPDIR'])


def _clean_dummy_atom_arrays(
    atomic_numbers: np.ndarray, positions: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Clean dummy atoms from the molecule.
    """
    non_dummy_inds = np.where(atomic_numbers != 0)[0]
    return atomic_numbers[non_dummy_inds], positions[non_dummy_inds]


def _clean_dummy_pharm_arrays(
    pharm_types: np.ndarray, pharm_ancs: np.ndarray, pharm_vecs: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Clean dummy pharmacophores from the molecule.
    """
    non_dummy_inds = np.where(pharm_types != P_TYPES.index('Dummy'))[0]
    return pharm_types[non_dummy_inds], pharm_ancs[non_dummy_inds], pharm_vecs[non_dummy_inds]


class ConfEval:
    """ Generated conformer evaluation pipeline """

    def __init__(self,
                 atoms: np.ndarray,
                 positions: np.ndarray,
                 solvent: Optional[str] = None,
                 num_processes: int = 1):
        """
        Base class for evaluation of a single generated conformer.

        Checks validity with RDKit pipeline and xTB single point calculation and optimization.
        Calculates 2D graph properties for valid molecules.

        Automatically aligns relaxed structure to the original structure via rdkit RMS.

        Arguments
        ---------
        atoms : np.ndarray (N,) of atomic numbers of the generated molecule or (N,M) one-hot
            encoding.
        positions : np.ndarray (N,3) of coordinates for the generated molecule's atoms.
        solvent : str solvent type for xtb relaxation
        num_processes : int (default = 1) number of processors to use for xtb relaxation and RDKit
            RMSD alignment.
        """
        self.xyz_block = None
        self.mol = None
        self.smiles = None
        self.molblock = None
        self.energy = None
        self.partial_charges = None

        self.solvent = solvent
        self.charge = 0

        self.xyz_block_post_opt = None
        self.mol_post_opt = None
        self.smiles_post_opt = None
        self.molblock_post_opt = None
        self.energy_post_opt = None
        self.partial_charges_post_opt = None

        self.is_valid = False
        self.is_valid_post_opt = False
        self.is_graph_consistent = False

        # 2D graph features
        self.SA_score = None
        self.QED = None
        self.logP = None
        self.fsp3 = None
        self.morgan_fp = None

        self.SA_score_post_opt = None
        self.QED_post_opt = None
        self.logP_post_opt = None
        self.fsp3_post_opt = None
        self.morgan_fp_post_opt = None

        # Consistency in 3D
        self.strain_energy = None
        self.rmsd = None

        # 1. Converts coords + atom_ids -> xyz block
        atoms, positions = _clean_dummy_atom_arrays(atoms, positions)
        # 2. Get mol from xyz block
        self.mol, self.charge, self.xyz_block = get_mol_from_atom_pos(atoms=atoms, positions=positions)

        # 3. Get xtb energy and charges of initial conformation
        try:
            self.energy, self.partial_charges = single_point_xtb_from_xyz(xyz_block=self.xyz_block,
                                                                          solvent=self.solvent,
                                                                          charge=self.charge,
                                                                          num_cores=num_processes,
                                                                          temp_dir=TMPDIR)
            self.partial_charges = np.array(self.partial_charges)
        except Exception:
            pass
        self.is_valid = self.mol is not None and self.partial_charges is not None
        if self.is_valid:
            self.smiles = Chem.MolToSmiles(Chem.RemoveHs(self.mol))
            self.molblock = Chem.MolToMolBlock(self.mol)

        # 4. Relax structure with xtb
        try:
            xtb_out = optimize_conformer_with_xtb_from_xyz_block(self.xyz_block,
                                                                solvent=self.solvent,
                                                                num_cores=num_processes,
                                                                charge=self.charge,
                                                                temp_dir=TMPDIR)
            self.xyz_block_post_opt, self.energy_post_opt, self.partial_charges_post_opt = xtb_out
            self.partial_charges_post_opt = np.array(self.partial_charges_post_opt)

            # 5. Check if relaxed_structure is valid
            self.mol_post_opt = extract_mol_from_xyz_block(xyz_block=self.xyz_block_post_opt,
                                                           charge=self.charge)
        except Exception:
            pass

        self.is_valid_post_opt = self.mol_post_opt is not None and self.partial_charges_post_opt is not None

        # 6. Check if 2D molecular graphs are consistent
        if self.is_valid and self.is_valid_post_opt:
            self.is_graph_consistent = Chem.MolToSmiles(self.mol) == Chem.MolToSmiles(self.mol_post_opt)
            # Align post-opt mol with RMSD
            mol_atom_ids = list(range(self.mol.GetNumAtoms()))
            mol_post_opt_atom_ids = list(range(self.mol_post_opt.GetNumAtoms()))
            AlignMol(prbMol=self.mol_post_opt, refMol=self.mol, atomMap=[i for i in zip(mol_post_opt_atom_ids, mol_atom_ids)])

        if self.is_valid_post_opt:
            self.smiles_post_opt = Chem.MolToSmiles(Chem.RemoveHs(self.mol_post_opt))
            self.molblock_post_opt = Chem.MolToMolBlock(self.mol_post_opt)

        # 7. Calculate strain energy with relaxed structure
        if self.energy is not None and self.energy_post_opt is not None:
            self.strain_energy = self.energy - self.energy_post_opt

        # 8. Calculate RMSD from relaxed structure
        if self.is_graph_consistent:
            mol_copy = deepcopy(Chem.RemoveHs(self.mol))
            mol_post_opt_copy = deepcopy(Chem.RemoveHs(self.mol_post_opt))
            self.rmsd = GetBestRMS(prbMol=mol_copy, refMol=mol_post_opt_copy, numThreads=num_processes)

        # 9. 2D graph properties
        if self.is_valid:
            self.SA_score = sascorer.calculateScore(Chem.RemoveHs(self.mol))
            self.QED = QED.qed(self.mol)
            self.logP = Crippen.MolLogP(self.mol)
            self.fsp3 = Lipinski.FractionCSP3(self.mol)
            self.morgan_fp = morgan_fp_gen.GetFingerprint(mol=Chem.RemoveHs(self.mol))

        # 10. 2D graph properties post optimization
        if self.is_valid_post_opt:
            self.SA_score_post_opt = sascorer.calculateScore(Chem.RemoveHs(self.mol_post_opt))
            self.QED_post_opt = QED.qed(self.mol_post_opt)
            self.logP_post_opt = Crippen.MolLogP(self.mol_post_opt)
            self.fsp3_post_opt = Lipinski.FractionCSP3(self.mol_post_opt)
            self.morgan_fp_post_opt = morgan_fp_gen.GetFingerprint(mol=Chem.RemoveHs(self.mol_post_opt))


    def to_pandas(self):
        """
        Convert the stored attributes to a pd.Series (for global attributes).

        Arguments
        ---------
        self

        Returns
        -------
        pd.Series : holds attributes in an easy to visualize way
        """
        global_attrs = {} # Global attributes

        for key, value in self.__dict__.items():
            global_attrs[key] = value

        series_global = pd.Series(global_attrs)

        return series_global


class ConsistencyEval(ConfEval):
    """
    Evaluation of the consistency between jointly generated molecules' features.
    Consistency in terms of similarity scores.
    """
    def __init__(self,
                 atoms: np.ndarray,
                 positions: np.ndarray,
                 surf_points: Optional[np.ndarray] = None,
                 surf_esp: Optional[np.ndarray] = None,
                 pharm_feats: Optional[Tuple[np.ndarray, np.ndarray, np.ndarray]] = None,
                 pharm_multi_vector: Optional[bool] = None,
                 solvent: Optional[str] = None,
                 probe_radius: float = 1.2,
                 num_processes: int = 1):
        """
        Consistency evaluation class for jointly generated molecule and features.

        Uses 3D similarity scoring functions. Inherits from ConfEval so that it
        can first run a conformer evaluation on the generated molecule.

        Must supply ``atoms`` and ``positions`` AND at least one of the features necessary for
        similarity scoring.

        Notes
        -----
        Important assumptions:

        - Gaussian width parameter (alpha) for surface similarity was fitted to a probe
          radius of 1.2 A.
        - ESP weighting parameter (lam) for electrostatic similarity is set to 0.3 which
          was tested for the above assumption.

        Parameters
        ----------
        atoms : np.ndarray
            Array of shape (N,) of atomic numbers of the generated molecule or (N, M)
            one-hot encoding.
        positions : np.ndarray
            Array of shape (N, 3) of coordinates for the generated molecule's atoms.
        surf_points : np.ndarray, optional
            Array of shape (M, 3) of generated surface point cloud.
        surf_esp : np.ndarray, optional
            Array of shape (M,) of generated electrostatic potential on surface.
        pharm_feats : tuple, optional
            Tuple of (pharm_types, pharm_ancs, pharm_vecs) where pharm_types is (P,)
            type of pharmacophore defined by shepherd_score.score.constants.P_TYPES,
            pharm_ancs is (P, 3) anchor positions, and pharm_vecs is (P, 3) unit vectors
            relative to anchor.
        pharm_multi_vector : bool, optional
            Use multiple vectors to represent Aro/HBA/HBD or single.
        solvent : str, optional
            Solvent type for xTB relaxation.
        probe_radius : float, optional
            Radius of probe atom used to generate solvent accessible surface.
            Default is 1.2 (vdW radius of hydrogen).
        num_processes : int, optional
            Number of processors to use for xTB relaxation. Default is 1.
        """
        if not (isinstance(atoms, np.ndarray) or isinstance(positions, np.ndarray)):
            raise ValueError(f"Must provide `atoms` and `positions` as np.ndarrays. Instead {type(atoms)} and {type(positions)} were given.")

        super().__init__(atoms=atoms, positions=positions, solvent=solvent, num_processes=num_processes)

        self.molec = None
        self.probe_radius = probe_radius
        self.molec_regen = None
        self.molec_post_opt = None

        self.sim_surf_consistent = None
        self.sim_esp_consistent = None
        self.sim_pharm_consistent = None

        self.sim_surf_consistent_relax = None
        self.sim_esp_consistent_relax = None
        self.sim_pharm_consistent_relax = None

        self.sim_surf_consistent_relax_optimal = None
        self.sim_esp_consistent_relax_optimal = None
        self.sim_pharm_consistent_relax_optimal = None

        if pharm_feats is not None:
            pharm_types, pharm_ancs, pharm_vecs = pharm_feats
            num_pharms = len(pharm_types)

            if pharm_ancs.shape != (num_pharms, 3) or pharm_vecs.shape != (num_pharms, 3):
                raise ValueError(
                    f'Provided pharmacophore features do not match dimensions: pharm_types {pharm_types.shape}, pharm_ancs {pharm_ancs.shape}, pharm_vecs {pharm_vecs.shape}'
                )
            pharm_types, pharm_ancs, pharm_vecs = _clean_dummy_pharm_arrays(pharm_types, pharm_ancs, pharm_vecs)
        else:
            pharm_types, pharm_ancs, pharm_vecs = None, None, None

        has_pharm_features = (isinstance(pharm_types, np.ndarray)
                              and isinstance(pharm_ancs, np.ndarray)
                              and isinstance(pharm_vecs, np.ndarray))
        if has_pharm_features and pharm_multi_vector is None:
            print('WARNING: Generated pharmacophore features provided, but `pharm_multi_vector` is None.')
            print('         Pharmacophore similarity not computed.')
        if not isinstance(surf_points, np.ndarray) and not isinstance(surf_esp, np.ndarray) and not has_pharm_features:
            raise ValueError('Must provide at least one of the generated representations: surface, electrostatics, or pharmacophores.')

        # Scoring parameters
        self.num_surf_points = len(surf_points) if surf_points is not None else None
        # Assumes no radius scaling with probe_radius=1.2
        self.alpha = ALPHA(self.num_surf_points) if self.num_surf_points is not None else None
        self.lam = 0.3 # Optimal lambda for probe_radius=1.2 -> ONLY TO BE USED FOR ESP ALIGNMENT
        self.lam_scaled = self.lam * LAM_SCALING # -> ONLY TO BE USED FOR get_overlap_esp*

        # Self-consistency of features for generated molecule
        if self.is_valid:
            # Generate a Molecule object with generated features
            self.molec = Molecule(
                self.mol,
                partial_charges=np.array(self.partial_charges),
                surface_points=surf_points,
                electrostatics=surf_esp,
                pharm_types=pharm_types,
                pharm_ancs=pharm_ancs,
                pharm_vecs=pharm_vecs,
                probe_radius=self.probe_radius
            )
            # Generate a Molecule object with regenerated features
            self.molec_regen = Molecule(
                self.mol,
                num_surf_points = self.num_surf_points,
                probe_radius=self.probe_radius,
                partial_charges = np.array(self.partial_charges),
                pharm_multi_vector = pharm_multi_vector if has_pharm_features else None
            )

            if self.molec.surf_pos is not None:
                self.sim_surf_consistent = get_overlap_np(
                    self.molec.surf_pos,
                    self.molec_regen.surf_pos,
                    alpha=self.alpha
                )

            if self.molec.surf_esp is not None and self.molec.surf_pos is not None:
                self.sim_esp_consistent = get_overlap_esp_np(
                    self.molec.surf_pos, self.molec_regen.surf_pos,
                    self.molec.surf_esp, self.molec_regen.surf_esp,
                    alpha=self.alpha,
                    lam=self.lam_scaled
                )

            if has_pharm_features and pharm_multi_vector is not None:
                self.sim_pharm_consistent = get_overlap_pharm_np(
                    self.molec.pharm_types,
                    self.molec_regen.pharm_types,
                    self.molec.pharm_ancs,
                    self.molec_regen.pharm_ancs,
                    self.molec.pharm_vecs,
                    self.molec_regen.pharm_vecs,
                    similarity='tanimoto',
                    extended_points=False,
                    only_extended=False
                )

        # Consistency between generated molecule and relaxed structure and features
        if self.is_valid and self.is_valid_post_opt:
            # Generate a Molecule object of relaxed structure
            self.molec_post_opt = Molecule(
                self.mol_post_opt,
                num_surf_points = self.num_surf_points,
                probe_radius=self.probe_radius,
                partial_charges = np.array(self.partial_charges_post_opt),
                pharm_multi_vector = pharm_multi_vector if has_pharm_features else None
            )

            # Score only since we already align w.r.t. RMS of the generated atomic point cloud
            if self.molec_post_opt.surf_pos is not None:
                self.sim_surf_consistent_relax = get_overlap_np(
                    self.molec.surf_pos,
                    self.molec_post_opt.surf_pos,
                    alpha=self.alpha
                )
            if self.molec_post_opt.surf_pos is not None and self.molec_post_opt.surf_esp is not None:
                self.sim_esp_consistent_relax = get_overlap_esp_np(
                    self.molec.surf_pos, self.molec_post_opt.surf_pos,
                    self.molec.surf_esp, self.molec_post_opt.surf_esp,
                    alpha=self.alpha,
                    lam=self.lam_scaled
                )
            if isinstance(pharm_multi_vector, bool) and self.molec_post_opt.pharm_ancs is not None and self.molec.pharm_ancs is not None:
                self.sim_pharm_consistent_relax = get_overlap_pharm_np(
                    self.molec.pharm_types,
                    self.molec_post_opt.pharm_types,
                    self.molec.pharm_ancs,
                    self.molec_post_opt.pharm_ancs,
                    self.molec.pharm_vecs,
                    self.molec_post_opt.pharm_vecs,
                    similarity='tanimoto',
                    extended_points=False,
                    only_extended=False
                )

            # Alignment with scoring functions
            mp_ref_and_relaxed = MoleculePair(self.molec,
                                              self.molec_post_opt,
                                              num_surf_points=self.num_surf_points,
                                              do_center=False)
            if self.molec_post_opt.surf_pos is not None:
                self.sim_surf_consistent_relax_optimal = self._align_with_surface(mp_ref_and_relaxed=mp_ref_and_relaxed)
            if self.molec_post_opt.surf_pos is not None and self.molec_post_opt.surf_esp is not None:
                self.sim_esp_consistent_relax_optimal = self._align_with_esp(mp_ref_and_relaxed=mp_ref_and_relaxed)
            if isinstance(pharm_multi_vector, bool) and self.molec_post_opt.pharm_ancs is not None and self.molec.pharm_ancs is not None:
                self.sim_pharm_consistent_relax_optimal = self._align_with_pharm(mp_ref_and_relaxed=mp_ref_and_relaxed)


    def _align_with_surface(self, mp_ref_and_relaxed: MoleculePair) -> float:
        """
        Align relaxed molecule to reference/target molecule with surface.

        Returns
        -------
        float : Surface similarity score of optimally aligned molecule.
        """
        _ = mp_ref_and_relaxed.align_with_surf(
            self.alpha,
            num_repeats=1,
            trans_init=False,
            use_jax=False
        )
        surf_similarity = mp_ref_and_relaxed.sim_aligned_surf
        return float(surf_similarity)


    def _align_with_esp(self, mp_ref_and_relaxed: MoleculePair) -> float:
        """
        Align relaxed molecule to reference/target molecule with ESP

        Returns
        -------
        float : ESP similarity score of optimally aligned molecule.
        """
        _ = mp_ref_and_relaxed.align_with_esp(
            self.alpha,
            lam=self.lam,
            num_repeats=1,
            trans_init=False,
            use_jax=False
        )
        esp_similarity = mp_ref_and_relaxed.sim_aligned_esp
        return float(esp_similarity)


    def _align_with_pharm(self, mp_ref_and_relaxed: MoleculePair) -> float:
        """
        Align relaxed molecule to reference/target molecule with pharmacophores

        Returns
        -------
        float : Pharmacophore similarity score of optimally aligned molecule.
        """
        aligned_fit_anchors, aligned_vectors = mp_ref_and_relaxed.align_with_pharm(
            similarity='tanimoto',
            extended_points=False,
            only_extended=False,
            num_repeats=1,
            trans_init=False,
            use_jax=False
        )
        pharm_similarity = mp_ref_and_relaxed.sim_aligned_pharm
        return float(pharm_similarity)


class ConditionalEval(ConfEval):
    """ Evaluation of conditionally generated molecules' quality and similarity. """

    def __init__(self,
                 ref_molec: Molecule,
                 atoms: np.ndarray,
                 positions: np.ndarray,
                 condition: str,
                 num_surf_points: int = 400,
                 pharm_multi_vector: Optional[bool] = None,
                 solvent: Optional[str] = None,
                 num_processes: int = 1):
        """
        Evaluation pipeline for conditionally-generated molecules.

        Inherits from ConfEval so that it can first run a conformer evaluation on the
        generated molecule.

        Notes
        -----
        Important assumptions:

        - Gaussian width parameter (alpha) for surface similarity assumes a probe radius
          of 1.2A.
        - ESP weighting parameter (lam) for electrostatic similarity is set to 0.3 which
          was tested for the above assumption.

        Parameters
        ----------
        ref_molec : Molecule
            Molecule object of reference/target molecule. Must contain the representation
            that was used for conditioning.
        atoms : np.ndarray
            Array of shape (N,) of atomic numbers of the generated molecule or (N, M)
            one-hot encoding.
        positions : np.ndarray
            Array of shape (N, 3) of coordinates for the generated molecule's atoms.
        condition : str
            Condition that the molecule was conditioned on. One of 'surface', 'esp',
            'pharm', or 'all'. Used for alignment. Choose 'esp' or 'all' if you want
            to compute ESP-aligned scores for other profiles.
        num_surf_points : int, optional
            Number of surface points to sample for similarity scoring. Default is 400.
        pharm_multi_vector : bool, optional
            Use multiple vectors to represent Aro/HBA/HBD or single.
        solvent : str, optional
            Solvent type for xTB relaxation.
        num_processes : int, optional
            Number of processors to use for xTB relaxation. Default is 1.
        """
        condition = condition.lower()
        self.condition = None
        if 'surf' in condition or 'shape' in condition:
            self.condition = 'surface'
        elif 'esp' in condition or 'electrostatic' in condition:
            self.condition = 'esp'
        elif 'pharm' in condition:
            self.condition = 'pharm'
        elif condition == 'all':
            self.condition = 'all'
        else:
            raise ValueError(f'`condition` must contain one of the following: "surf", "esp", "pharm", or "all". Instead, {condition} was given.')

        super().__init__(atoms=atoms,
                         positions=positions,
                         solvent=solvent,
                         num_processes=num_processes)

        self.sim_surf_target = None
        self.sim_esp_target = None
        self.sim_pharm_target = None

        self.sim_surf_target_relax = None
        self.sim_esp_target_relax = None
        self.sim_pharm_target_relax = None

        self.sim_surf_target_relax_optimal = None
        self.sim_esp_target_relax_optimal = None
        self.sim_pharm_target_relax_optimal = None

        self.sim_surf_target_relax_esp_aligned = None
        self.sim_pharm_target_relax_esp_aligned = None

        # Scoring parameters
        self.num_surf_points = num_surf_points
        self.alpha = ALPHA(self.num_surf_points) # Fitted to probe_radius=1.2
        self.lam = 0.3 # Optimal lambda for probe_radius=1.2 -> ONLY TO BE USED FOR ESP ALIGNMENT
        self.lam_scaled = self.lam * LAM_SCALING # -> ONLY TO BE USED FOR get_overlap_esp*

        self.ref_molec = ref_molec # Reference Molecule object
        self.molec_regen = None
        self.molec_post_opt = None

        if self.is_valid:
            # Generate a Molecule object with regenerated features
            self.molec_regen = Molecule(
                self.mol,
                num_surf_points = self.num_surf_points,
                partial_charges = np.array(self.partial_charges),
                pharm_multi_vector = pharm_multi_vector,
                probe_radius=self.ref_molec.probe_radius
            )

            if self.molec_regen.surf_pos is not None:
                self.sim_surf_target = get_overlap_np(
                    self.molec_regen.surf_pos,
                    self.ref_molec.surf_pos,
                    alpha=self.alpha
                )

            if self.molec_regen.surf_esp is not None and self.molec_regen.surf_pos is not None:
                self.sim_esp_target = get_overlap_esp_np(
                    self.molec_regen.surf_pos, self.ref_molec.surf_pos,
                    self.molec_regen.surf_esp, self.ref_molec.surf_esp,
                    alpha=self.alpha,
                    lam=self.lam_scaled
                )

            if pharm_multi_vector is not None and self.ref_molec.pharm_ancs is not None:
                self.sim_pharm_target = get_overlap_pharm_np(
                    self.molec_regen.pharm_types,
                    self.ref_molec.pharm_types,
                    self.molec_regen.pharm_ancs,
                    self.ref_molec.pharm_ancs,
                    self.molec_regen.pharm_vecs,
                    self.ref_molec.pharm_vecs,
                    similarity='tanimoto',
                    extended_points=False,
                    only_extended=False
                )

        # Similarity between relaxed structure and target molecule -> first align
        if self.is_valid_post_opt:
            # Generate a Molecule object of relaxed structure
            self.molec_post_opt = Molecule(
                self.mol_post_opt,
                num_surf_points = self.num_surf_points,
                partial_charges = np.array(self.partial_charges_post_opt),
                pharm_multi_vector = pharm_multi_vector,
                probe_radius=self.ref_molec.probe_radius
            )

            # Score based on RMS alignment
            if self.molec_post_opt.surf_pos is not None:
                self.sim_surf_target_relax = get_overlap_np(
                    self.molec_post_opt.surf_pos,
                    self.ref_molec.surf_pos,
                    alpha=self.alpha
                )

            if self.molec_post_opt.surf_esp is not None and self.molec_post_opt.surf_pos is not None:
                self.sim_esp_target_relax = get_overlap_esp_np(
                    self.molec_post_opt.surf_pos, self.ref_molec.surf_pos,
                    self.molec_post_opt.surf_esp, self.ref_molec.surf_esp,
                    alpha=self.alpha,
                    lam=self.lam_scaled
                )

            if pharm_multi_vector is not None and self.ref_molec.pharm_ancs is not None:
                self.sim_pharm_target_relax = get_overlap_pharm_np(
                    self.molec_post_opt.pharm_types,
                    self.ref_molec.pharm_types,
                    self.molec_post_opt.pharm_ancs,
                    self.ref_molec.pharm_ancs,
                    self.molec_post_opt.pharm_vecs,
                    self.ref_molec.pharm_vecs,
                    similarity='tanimoto',
                    extended_points=False,
                    only_extended=False
                )

            # Align and score w.r.t. specified condition
            mp_ref_and_relaxed = MoleculePair(self.ref_molec,
                                              self.molec_post_opt,
                                              num_surf_points=self.num_surf_points,
                                              do_center=False)
            if (self.condition == 'surface' or self.condition == 'all') and self.molec_post_opt.surf_pos is not None:
                self.sim_surf_target_relax_optimal = self._align_with_surface(mp_ref_and_relaxed=mp_ref_and_relaxed)
            if (self.condition == 'esp' or self.condition == 'all') and self.molec_post_opt.surf_pos is not None and self.molec_post_opt.surf_esp is not None:
                self.sim_esp_target_relax_optimal = self._align_with_esp(mp_ref_and_relaxed=mp_ref_and_relaxed)
            if (self.condition == 'pharm' or self.condition == 'all') and isinstance(pharm_multi_vector, bool):
                self.sim_pharm_target_relax_optimal = self._align_with_pharm(mp_ref_and_relaxed=mp_ref_and_relaxed)

            # Compute ESP-aligned surf and pharmacophore similarity scores
            if mp_ref_and_relaxed.transform_esp is not None and self.condition in ('esp', 'all'):
                molec_post_opt_esp_aligned = mp_ref_and_relaxed.get_transformed_molecule(mp_ref_and_relaxed.transform_esp)
                esp_aligned_molec_pair = MoleculePair(ref_mol=self.ref_molec,
                                                      fit_mol=molec_post_opt_esp_aligned,
                                                      num_surf_points=self.ref_molec.num_surf_points,
                                                      do_center=False)
                self.sim_surf_target_relax_esp_aligned = esp_aligned_molec_pair.score_with_surf(
                    alpha=self.alpha, use='np'
                )
                if isinstance(pharm_multi_vector, bool):
                    self.sim_pharm_target_relax_esp_aligned = esp_aligned_molec_pair.score_with_pharm(
                        similarity='tanimoto',
                        extended_points=False,
                        only_extended=False,
                        use='np'
                    )


    def _align_with_surface(self, mp_ref_and_relaxed: MoleculePair) -> float:
        """
        Align relaxed molecule to reference/target molecule with surface.

        Returns
        -------
        float : Surface similarity score of optimally aligned molecule.
        """
        _ = mp_ref_and_relaxed.align_with_surf(
            self.alpha,
            num_repeats=1,
            trans_init=False,
            use_jax=False
        )

        surf_similarity = mp_ref_and_relaxed.sim_aligned_surf
        return float(surf_similarity)


    def _align_with_esp(self, mp_ref_and_relaxed: MoleculePair) -> float:
        """
        Align relaxed molecule to reference/target molecule with ESP

        Returns
        -------
        float : ESP similarity score of optimally aligned molecule.
        """
        _ = mp_ref_and_relaxed.align_with_esp(
            self.alpha,
            lam=self.lam,
            num_repeats=1,
            trans_init=False,
            use_jax=False
        )
        esp_similarity = mp_ref_and_relaxed.sim_aligned_esp
        return float(esp_similarity)


    def _align_with_pharm(self, mp_ref_and_relaxed: MoleculePair) -> float:
        """
        Align relaxed molecule to reference/target molecule with pharmacophores

        Returns
        -------
        float : Pharmacophore similarity score of optimally aligned molecule.
        """
        aligned_fit_anchors, aligned_vectors = mp_ref_and_relaxed.align_with_pharm(
            similarity='tanimoto',
            extended_points=False,
            only_extended=False,
            num_repeats=1,
            trans_init=False,
            use_jax=False
        )
        pharm_similarity = mp_ref_and_relaxed.sim_aligned_pharm
        return float(pharm_similarity)



"""
Script to run conditional evaluation of generated molecules.
"""

import os
from typing import Optional
from pathlib import Path
import pickle
import argparse
import open3d

import numpy as np
import pandas as pd
from rdkit import Chem

from shepherd_score.score.constants import COULOMB_SCALING
from shepherd_score.container import Molecule, MoleculePair
from shepherd_score.generate_point_cloud import get_atomic_vdw_radii, get_molecular_surface
from shepherd_score.generate_point_cloud import get_electrostatics_given_point_charges
from shepherd_score.conformer_generation import charges_from_single_point_conformer_with_xtb

from shepherd_score.evaluations.evaluate import ConditionalEvalPipeline

if 'TMPDIR' in os.environ:
    TMPDIR = os.environ['TMPDIR']
else:
    TMPDIR = './'


def run_conditional_eval(sample_id,
                         job_id,
                         num_tasks,
                         load_file_path,
                         solvent: Optional[str] = None,
                         num_processes: int = 1,
                         probe_radius: float = 0.6
                         ) -> None:
    """
    Run conditional evaluation and save. Split it jobs by the samples file and the job id.
    """
    load_file_path_ = Path(load_file_path)
    if not load_file_path_.is_file():
        raise ValueError(f'Provided path is not a file: {load_file_path_}')

    load_dir_path = load_file_path_.parent
    if not load_dir_path.is_dir():
        raise ValueError(f'Provided path is not a directory: {load_dir_path}')
    save_file_dir = load_dir_path / f'cond_eval_sample_{sample_id}'
    save_file_dir.mkdir(parents=True, exist_ok=True)

    with open(load_file_path_, 'rb') as f:
        samples = pickle.load(f)

    ref_mol = Chem.MolFromMolBlock(samples[0], removeHs=False)
    ref_partial_charges = samples[1]
    surface_points = samples[2] # noqa: F841
    electrostatics = samples[3] # noqa: F841
    pharm_types = samples[4]
    pharm_ancs = samples[5]
    pharm_vecs = samples[6]
    ref_molec = Molecule(ref_mol,
                         probe_radius=probe_radius,
                         partial_charges=np.array(ref_partial_charges),
                         num_surf_points=400,
                         pharm_multi_vector=False,
                         pharm_types=pharm_types,
                         pharm_ancs=pharm_ancs,
                         pharm_vecs=pharm_vecs)

    generated_mols = [(samples[-1][i]['x1']['atoms'], samples[-1][i]['x1']['positions']) for i in range(len(samples[-1]))]

    subselected_gen_mols = generated_mols[job_id:len(generated_mols):num_tasks]

    print(f'Starting Conditional Eval Pipeline on sample {sample_id} and job {job_id}')
    cond_pipe = ConditionalEvalPipeline(
        ref_molec=ref_molec,
        generated_mols=subselected_gen_mols,
        condition='all',
        num_surf_points=400,
        pharm_multi_vector=False,
        solvent=solvent
    )

    # Run evaluation
    cond_pipe.evaluate(num_processes=num_processes, verbose=True)

    save_global_series, save_local_df = cond_pipe.to_pandas()

    save_file_global = save_file_dir / f'cond_eval_global_{job_id}.pkl'
    save_global_series.to_pickle(save_file_global)
    save_file_local = save_file_dir / f'cond_eval_local_{job_id}.pkl'
    save_local_df.to_pickle(save_file_local)

    print(f'Finished {job_id} evaluation!\nSaved global attributes to: {save_file_global}')
    print(f'Saved local attributes to: {save_file_local}')


def run_conditional_eval_by_sample_only(
        sample_id,
        load_dir,
        conditioning_type: str,
        solvent: Optional[str] = None,
        num_processes: int = 1,
        probe_radius: float = 0.6
        ) -> None:
    """
    Run conditional evaluation and save. Do not split up each samples file by job id.
    """
    load_dir_path = Path(load_dir) / conditioning_type
    if not load_dir_path.is_dir():
        raise ValueError(f'Provided path is not a directory: {load_dir_path}')

    if conditioning_type == 'x2':
        condition = 'surf'
    elif conditioning_type == 'x3':
        condition = 'esp'
    elif conditioning_type == 'x4':
        condition = 'pharm'

    save_file_dir = load_dir_path / 'cond_evals'
    save_file_dir.mkdir(parents=True, exist_ok=True)

    load_file_path = load_dir_path / f'samples_{sample_id}.pickle'
    with open(load_file_path, 'rb') as f:
        samples = pickle.load(f)

    ref_mol = Chem.MolFromMolBlock(samples[0], removeHs=False)
    ref_partial_charges = samples[1]
    surface_points = samples[2] # noqa: F841
    electrostatics = samples[3] # noqa: F841
    pharm_types = samples[4]
    pharm_ancs = samples[5]
    pharm_vecs = samples[6]
    ref_molec = Molecule(ref_mol,
                         probe_radius=probe_radius,
                         partial_charges=np.array(ref_partial_charges),
                         num_surf_points=400,
                         pharm_multi_vector=False,
                         pharm_types=pharm_types,
                         pharm_ancs=pharm_ancs,
                         pharm_vecs=pharm_vecs)

    generated_mols = [(samples[-1][i]['x1']['atoms'], samples[-1][i]['x1']['positions']) for i in range(len(samples[-1]))]

    print(f'Starting Conditional Eval Pipeline on sample {sample_id}.')
    cond_pipe = ConditionalEvalPipeline(
        ref_molec=ref_molec,
        generated_mols=generated_mols,
        condition=condition,
        num_surf_points=400,
        pharm_multi_vector=False,
        solvent=solvent
    )

    # Run evaluation
    cond_pipe.evaluate(num_processes=num_processes, verbose=True)

    save_global_series, save_local_df = cond_pipe.to_pandas()

    save_file_global = save_file_dir / 'cond_eval_global.pkl'
    save_global_series.to_pickle(save_file_global)
    save_file_local = save_file_dir / 'cond_eval_local.pkl'
    save_local_df.to_pickle(save_file_local)

    print(f'Finished evaluation!\nSaved global attributes to: {save_file_global}')
    print(f'Saved local attributes to: {save_file_local}')


def run_conditional_eval_frag(job_id,
                              num_tasks,
                              load_file,
                              solvent: Optional[str] = None,
                              num_processes: int = 1,
                              probe_radius: float = 0.6
                              ) -> None:
    """
    Run conditional evaluation and save. Split it jobs by the samples file and the job id.
    """
    print('Starting Frag')
    load_file_path = Path(load_file)
    load_dir_path = load_file_path.parent
    if not load_dir_path.is_dir():
        raise ValueError(f'Provided path is not a directory: {load_dir_path}')
    save_file_dir = load_dir_path / 'cond_evals'
    save_file_dir.mkdir(parents=True, exist_ok=True)

    with open(load_file_path, 'rb') as f:
        samples = pickle.load(f)

    # Create reference mol from all fragments
    mols = [Chem.MolFromMolBlock(s, removeHs=False) for s in samples[0]]
    centers = []
    radii = []
    for mol in mols:
        centers.append(mol.GetConformer().GetPositions())
        radii.append(get_atomic_vdw_radii(mol))
    centers_comb = np.concatenate(centers)
    radii_comb = np.concatenate(radii)
    surface_points = get_molecular_surface(centers_comb, radii_comb, num_points=400, probe_radius=probe_radius)

    partial_charges = []
    esps = []
    for i, mol in enumerate(mols):
        charges = charges_from_single_point_conformer_with_xtb(mol, solvent='water', num_cores=1, temp_dir=TMPDIR)
        partial_charges.append(charges)
        esps.append(get_electrostatics_given_point_charges(charges, centers[i], surface_points))
    avg_esp = np.stack(esps).mean(axis=0)

    print('Finished generating the merged reference Molecule object.')

    # just choose the first molblock, it doesn't affect anything once you make the Molecule object
    ref_mol = mols[0]
    ref_partial_charges = partial_charges[0]
    # surface_points = samples[2] ignore since len=75
    # electrostatics = samples[3] ignore since len=75
    pharm_types = samples[3]
    pharm_ancs = samples[4]
    pharm_vecs = samples[5]
    ref_molec = Molecule(ref_mol,
                         probe_radius=probe_radius,
                         partial_charges=np.array(ref_partial_charges),
                        #  num_surf_points=400,
                         surface_points=surface_points, # We generate the surface
                         electrostatics=avg_esp, # We generate the esp
                         pharm_multi_vector=False,
                         pharm_types=pharm_types,
                         pharm_ancs=pharm_ancs,
                         pharm_vecs=pharm_vecs)

    generated_mols = [(samples[-1][i]['x1']['atoms'], samples[-1][i]['x1']['positions']) for i in range(len(samples[-1]))]

    subselected_gen_mols = generated_mols[job_id:len(generated_mols):num_tasks]

    print(f'Starting Conditional Eval Pipeline on Fragment mergining and job {job_id}')
    cond_pipe = ConditionalEvalPipeline(
        ref_molec=ref_molec,
        generated_mols=subselected_gen_mols,
        condition='all',
        num_surf_points=400,
        pharm_multi_vector=False,
        solvent=solvent
    )

    # Run evaluation
    cond_pipe.evaluate(num_processes=num_processes, verbose=True)

    save_global_series, save_local_df = cond_pipe.to_pandas()

    save_file_global = save_file_dir / f'cond_eval_global_{job_id}.pkl'
    save_global_series.to_pickle(save_file_global)
    save_file_local = save_file_dir / f'cond_eval_local_{job_id}.pkl'
    save_local_df.to_pickle(save_file_local)

    print(f'Finished {job_id} evaluation!\nSaved global attributes to: {save_file_global}')
    print(f'Saved local attributes to: {save_file_local}')


if __name__=='__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--load-file-path', type=str, help='Path file with generated samples OR if GDB conditional samples the path to the directory.')
    parser.add_argument('--task-id', type=int, help='Task ID.')
    parser.add_argument('--num-tasks', type=int, help='Number of tasks.')
    parser.add_argument('--sample-id', type=int, required=True, help='Index used to load the file. [0,4]')
    parser.add_argument('--task', type=str, required=True, help='Choose from "NP", "frag", or "GDB"')
    args = parser.parse_args()
    print(args)

    file_path = Path(args.load_file_path)
    if not file_path.is_file():
        raise ValueError('Provided --load-file-path is not a directory.')
    my_task_id = int(args.task_id)
    num_tasks = int(args.num_tasks)
    sample_id = int(args.sample_id)
    task = str(args.task)

    print(f'Loading from {file_path}')
    if 'NP' in task:
        run_conditional_eval(
            sample_id=sample_id,
            job_id=my_task_id,
            num_tasks=num_tasks,
            load_file_path=file_path,
            solvent='water',
            num_processes=4,
            probe_radius=0.6
        )

    if 'frag' in task:
        run_conditional_eval_frag(
            job_id=my_task_id,
            num_tasks=num_tasks,
            load_file=file_path,
            solvent='water',
            num_processes=4,
            probe_radius=0.6
        )

    # total evals: 60-120
    if 'GDB' in task:
        samples_numbers = np.arange(100)
        samples_to_eval = samples_numbers[my_task_id:len(samples_numbers):num_tasks]
        print(f'Running GDB_conditional for these samples:\n{samples_to_eval}')
        # Go through every assigned sample id (1-2)
        for sample_idx in samples_to_eval:
            # Go through every representation (20*3)
            for conditioning in ('x2', 'x3', 'x4'):
                # Use task ID as sample id
                print(f'Running sample {sample_idx} for {conditioning} condition.')
                run_conditional_eval_by_sample_only(sample_id=sample_idx,
                                                    load_dir=file_path,
                                                    conditioning_type=conditioning,
                                                    solvent=None,
                                                    num_processes=4,
                                                    probe_radius=0.6)



"""
Module contains functions to generate point clouds of
molecular surfaces.

Surface generation workflow adapted from Open Drug Discovery Toolkit:
    https://oddt.readthedocs.io/en/latest/_modules/oddt/surface.html
"""
from typing import Tuple, List

import numpy as np
from shepherd_score.score.constants import COULOMB_SCALING
import open3d as o3d

import rdkit
from rdkit import Chem
from rdkit.Chem import AllChem
from scipy.spatial import distance
from typing import Union

PT = Chem.GetPeriodicTable()


def get_atom_coords(mol: rdkit.Chem.Mol,
                    MMFF_optimize: bool = True
                    ) -> Tuple[Chem.Mol, np.ndarray]:
    """
    Get the coordinates of all atoms in a molecule using rdkit.
    If the rdkit.Chem.mol object already has a conformer it just retrieves the coordinates
    without optimizing using MMFF.

    Parameters
    ----------
    mol : rdkit.Chem.Mol object
        RDKit molecule object

    MMFF_optimize : bool
        Whether or not to use MMFF to optimize geometry

    Returns
    -------
    tuple
        rdkit.Chem.Mol
            Mol object. If the input mol did not have a conformer, it uses MMFF to optimize an
            embeded molecule and includes hydrogens.
        np.ndarray: shape = (N,3)
            Positions of each atom's center.
    """
    try:
        mol.GetConformer()
    except ValueError:
        try:
            mol = Chem.AddHs(mol)
            Chem.AllChem.EmbedMolecule(mol, maxAttempts = 200)
            if MMFF_optimize:
                Chem.AllChem.MMFFOptimizeMolecule(mol)

            mol.GetConformer() # test whether conformer generation succeeded
        except Exception as e:
            print('Failed to embed molecule:', e)
            return None
    return mol, mol.GetConformer().GetPositions()


def get_atomic_vdw_radii(mol: rdkit.Chem.Mol) -> np.ndarray:
    """
    Get the van der Waals radii of all atoms in a molecule using rdkit.

    Parameters
    ----------
    mol : rdkit.Chem.Mol object

    Returns
    -------
    np.ndarray : shape = (N,)
        vdW radii for each atom.
    """
    radii = np.zeros((mol.GetNumAtoms(),))
    for i, _ in enumerate(radii):
        # get the van der Waals radii of each atom
        radii[i] = PT.GetRvdw(mol.GetAtomWithIdx(i).GetAtomicNum())
    return radii


###################################
# Sampling from molecular surface #
###################################
def sample_molecular_surface_with_radius(centers: np.ndarray,
                                         radii: Union[np.ndarray, List],
                                         probe_radius: float = 1.2,
                                         num_samples_per_atom: int = 20
                                         ) -> np.ndarray:
    """
    Samples points from the surface of vdW radius of atoms and combines it into one molecule (Vectorized).

    Parameters
    ----------
    centers : np.ndarray (N, 3)
        Cartesian coordinates of the atom centers of a molecule.

    radii : np.ndarray (N,)
        van der Waals radii of each atom (in Angstrom) in the same order as the centers parameter.

    probe_radius : float (default = 1.2)
        The radius of a probe atom to act as a "solvent accessible surface".
        Default = 1.2 angstroms which is the radius of a Hydrogen atom.

    num_samples_per_atom : int, optional
        Number of points to sample from the surface of each atom. Default is 20.
        Note that this value is scaled by a given atom's relative vdW radius to a carbon
        and SQUARED. Typically choose a value between 15 and 35. For example, if set to
        20, a carbon atom would have 400 sampled points.

    Returns
    -------
    np.ndarray
        Array of shape (N*num_points_per_atom, 3) containing the coordinates of each
        point sampled from each atom.
    """
    # get surface radius based on vdW radii, cutoff
    # cutoff = 1.4
    if num_samples_per_atom > 50:
        raise ValueError('Do not set num_samples_per_atom to be larger than 50 for performance\
                         issues. The number is squared internally.')
    if not isinstance(radii, np.ndarray):
        radii = np.array(radii)
    # surf_radii = cutoff * (radii / 1.7) + probe_radius # scaled vdW radius + probe radius
    surf_radii = radii + probe_radius # vdW radius + probe radius
    # get number of samples per atom dependent on vdw radii
    num_samples_per_atom = np.ceil((num_samples_per_atom * (radii / 1.7))**2)

    # Sample unit surfaces from normal distribution
    x = np.random.normal(size=(int(num_samples_per_atom.sum()), 3))
    # reformat radii and center arrays to match number of samples per atom
    Rs = np.repeat(surf_radii.reshape((-1,1)), [int(n) for n in num_samples_per_atom], axis=0)
    centers = np.repeat(centers, [int(n) for n in num_samples_per_atom], axis=0)

    # Scale surfaces by radii and translate to centers
    surface = ((x / np.linalg.norm(x, axis=1)[:, np.newaxis]) * Rs) + centers
    return surface


def sample_molecular_surface_with_radius_fibonacci(centers: np.ndarray,
                                                   radii: Union[np.ndarray, List],
                                                   probe_radius: float = 1.2,
                                                   num_samples_per_atom: int = 20
                                                   ) -> np.ndarray:
    """
    Samples points from the surface of vdW radius of atoms and combines it into one molecule (Vectorized).

    Parameters
    ----------
    centers : np.ndarray (N, 3)
        Cartesian coordinates of the atom centers of a molecule.

    radii : np.ndarray (N,)
        van der Waals radii of each atom (in Angstrom) in the same order as the centers parameter.

    probe_radius : float (default = 1.2)
        The radius of a probe atom to act as a "solvent accessible surface".
        Default = 1.2 angstroms which is the radius of a Hydrogen atom.

    num_samples_per_atom : int, optional
        Number of points to sample from the surface of each atom. Default is 20.
        Note that this value is scaled by a given atom's relative vdW radius to a carbon
        and SQUARED. Typically choose a value between 15 and 35. For example, if set to
        20, a carbon atom would have 400 sampled points.

    Returns
    -------
    np.ndarray
        Array of shape (N*num_points_per_atom, 3) containing the coordinates of each
        point sampled from each atom.
    """
    # get surface radius based on vdW radii, cutoff
    # cutoff = 1.4
    if num_samples_per_atom > 50:
        raise ValueError('Do not set num_samples_per_atom to be larger than 50 for performance\
                         issues. The number is squared internally.')
    if not isinstance(radii, np.ndarray):
        radii = np.array(radii)
    # surf_radii = cutoff * (radii / 1.7) + probe_radius # scaled vdW radius + probe radius
    surf_radii = radii + probe_radius # vdW radius + probe radius
    # get number of samples per atom dependent on vdw radii
    num_samples_per_atom = np.ceil((num_samples_per_atom * (radii / 1.7))**2)

    # store points for spheres generated by same sized radii since deterministic
    spheres = {num_samples : _get_points_fibonacci(num_samples) for num_samples in set(num_samples_per_atom)}
    # Apply a random SO(3) rotation to each atom's sphere seperately
    # spheres = np.vstack([np.dot(spheres[num_samples], special_ortho_group.rvs(3).T) for num_samples in num_samples_per_atom])
    # Don't apply random SO(3) rotation since mesh sampling is stochastic and 5-8x faster.
    spheres = np.vstack([spheres[num_samples] for num_samples in num_samples_per_atom])

    # reformat radii and center arrays to match number of samples per atom for elementwise mult and add
    Rs = np.repeat(surf_radii.reshape((-1,1)), [int(n) for n in num_samples_per_atom], axis=0)
    centers = np.repeat(centers, [int(n) for n in num_samples_per_atom], axis=0)

    # Scale surfaces by radii and translate to centers
    surface = (spheres * Rs) + centers
    return surface


def _get_points_fibonacci(num_samples: int):
    """
    Generate points on unit sphere using fibonacci approach.
    Adapted from Morfeus:
    https://github.com/digital-chemistry-laboratory/morfeus/blob/main/morfeus/geometry.py

    Parameters
    ----------
    num_samples : int
        Number of points to sample from the surface of a sphere

    Returns
    -------
    np.ndarray (num_samples,3)
        Coordinates of the sampled points.
    """
    offset = 2.0 / num_samples
    increment = np.pi * (3.0 - np.sqrt(5.0))

    i = np.arange(num_samples)
    y = ((i * offset) - 1) + (offset / 2)
    r = np.sqrt(1 - np.square(y))
    phi = np.mod((i + 1), num_samples) * increment
    x = np.cos(phi) * r
    z = np.sin(phi) * r

    points = np.column_stack((x, y, z))
    return points


def get_point_cloud(points: np.ndarray,
                    color: List[float] = [0.0, 0.0, 0.0]
                    ) -> o3d.geometry.PointCloud:
    """
    Convert np.ndarray of points to a Open3D Point Cloud object.

    Parameters
    ----------
    points : np.ndarray (N, 3)
        Coordinates of points.

    color : list[float] (default=[0,0,0] (black))

    Returns
    -------
    open3d.geometry.PointCloud object
    """
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)
    # pcd.paint_uniform_color(color)
    return pcd


def _get_molecular_surface_mesh(centers: np.ndarray,
                                radii: Union[np.ndarray, List],
                                num_samples_per_atom: int = 25,
                                probe_radius: float = 1.2,
                                ball_radii: List[float] = [1.2],
                                color: List[float] = [1.0, 0.0, 0.0]
                                ) -> Tuple[o3d.geometry.TriangleMesh, o3d.geometry.PointCloud]:
    """
    Generate a surface mesh representation of a molecule's surface. The dense point cloud is also
    returned.

    Parameters
    ----------
    centers : np.ndarray (N, 3)
        Cartesian coordinates of the atom centers of a molecule.

    radii : np.ndarray (N,)
        van der Waals radii of each atom (in Angstrom) in the same order as the centers parameter.

    num_samples_per_atom : int (default = 20)
        Number of points to sample from the surface of each atom.
        Note that this value is scaled by a given atom's relative vdW radius to a carbon and
        SQUARED. Typically choose a value between 15 and 35.
            E.g., if set to 20, a carbon atom would have 400 sampled points.

    probe_radius : float (default = 1.2)
        The radius of a probe atom to act as a "solvent accessible surface".
        Default = 1.2 angstroms which is the radius of a Hydrogen atom.

    ball_radii : list[float] (default = [1.2])
        The radius of the ball(s) used in Open3D's ball pivoting algorithm to generate a triangle
        mesh.

    color : list[float] (default = [1., 0., 0.])
        RGB color values for the point cloud (default is red).

    Returns
    -------
    tuple
        o3d.geometry.TriangleMesh : Mesh representing the molecular surface.
        o3d.geometry.PointCloud : Dense point cloud representing the molecular surface.
    """
    points = sample_molecular_surface_with_radius_fibonacci(centers=centers,
                                                            radii=radii,
                                                            probe_radius=probe_radius,
                                                            num_samples_per_atom=num_samples_per_atom
                                                            )
    # distances of every point with respect to the centers of each atom
    dist_matrix = distance.cdist(points, centers)
    # mask out the points within vdw radius of each atom
    mask = np.where(np.all(dist_matrix >= radii + probe_radius - 0.01, axis=1), 1., 0.).astype(bool)

    # generate point cloud
    pcd = get_point_cloud(points[mask], color=color)

    # Generate surface mesh and sample from it evenly
    pcd.estimate_normals()
    mesh = o3d.geometry.TriangleMesh.create_from_point_cloud_ball_pivoting(pcd, o3d.utility.DoubleVector(ball_radii))
    return mesh, points


def get_molecular_surface_point_cloud(centers: np.ndarray,
                                      radii: Union[np.ndarray, List],
                                      num_points: Union[int, None] = None,
                                      num_samples_per_atom: int = 25,
                                      probe_radius: float = 1.2,
                                      ball_radii: List[float] = [1.2],
                                      color: List[float] = [1.0, 0.0, 0.0]
                                      ) -> o3d.geometry.PointCloud:
    """
    Gets the point cloud representation of a molecule's van der Waals surface. Takes into account
    the vdW radii of different atoms. Removes overlapping points within vdW radii of neighboring
    atoms.

    Parameters
    ----------
    centers : np.ndarray (N, 3)
        Cartesian coordinates of the atom centers of a molecule.

    radii : np.ndarray (N,)
        van der Waals radii of each atom (in Angstrom) in the same order as the centers parameter.

    num_points : int (default = None)
        The total number of points in the final point cloud. If None, it returns as many as what
        was left after cleaning up the atom-sampled surface point cloud.

    num_samples_per_atom : int, optional
        Number of points to sample from the surface of each atom. Default is 25.
        Note that this value is scaled by a given atom's relative vdW radius to a carbon
        and SQUARED. Typically choose a value between 15 and 35. For example, if set to
        20, a carbon atom would have 400 sampled points.

    probe_radius : float, optional
        The radius of a probe atom to act as a "solvent accessible surface".
        Default is 1.2 angstroms which is the radius of a Hydrogen atom.

    ball_radii : list, optional
        The radius of the ball(s) used in Open3D's ball pivoting algorithm to generate
        a triangle mesh. Default is [1.2].

    color : list, optional
        RGB color values for the point cloud (default is [1., 0., 0.] red).

    Returns
    -------
    o3d.geometry.PointCloud
        Point cloud object representation of the molecular surface.
    """
    mesh, points = _get_molecular_surface_mesh(centers=centers,
                                               radii=radii,
                                               num_samples_per_atom=num_samples_per_atom,
                                               probe_radius=probe_radius,
                                               ball_radii=ball_radii,
                                               color=color)
    if num_points is None:
        num_points = len(points)
    pcd = mesh.sample_points_poisson_disk(num_points)
    return pcd


def get_molecular_surface(centers:np.ndarray,
                          radii:Union[np.ndarray, List],
                          num_points:Union[int, None] = None,
                          num_samples_per_atom:int = 25,
                          probe_radius: float = 1.2,
                          ball_radii: List[float] = [1.2],
                          ) -> np.ndarray:
    """
    Gets the point cloud representation of a molecule's van der Waals surface and outputs a
    numpy array. Takes into account the vdW radii of different atoms. Removes overlapping points
    within vdW radii of neighboring atoms.

    Parameters
    ----------
    centers : np.ndarray (N, 3)
        Cartesian coordinates of the atom centers of a molecule.

    radii : np.ndarray (N,)
        van der Waals radii of each atom (in Angstrom) in the same order as the centers parameter.

    num_points : int (default = None)
        The total number of points in the final point cloud. If None, it returns as many as what
        was left after cleaning up the atom-sampled surface point cloud.

    num_samples_per_atom : int, optional
        Number of points to sample from the surface of each atom. Default is 25.
        Note that this value is scaled by a given atom's relative vdW radius to a carbon
        and SQUARED. Typically choose a value between 15 and 35. For example, if set to
        20, a carbon atom would have 400 sampled points.

    probe_radius : float, optional
        The radius of a probe atom to act as a "solvent accessible surface".
        Default is 1.2 angstroms which is the radius of a Hydrogen atom.

    ball_radii : list, optional
        The radius of the ball(s) used in Open3D's ball pivoting algorithm to generate
        a triangle mesh. Default is [1.2].

    Returns
    -------
    np.ndarray
        Coordinates of points representing the molecular surface.
    """
    pcd = get_molecular_surface_point_cloud(centers=centers, radii=radii,
                                            num_points=num_points,
                                            num_samples_per_atom=num_samples_per_atom,
                                            probe_radius=probe_radius,
                                            ball_radii=ball_radii)
    return np.asarray(pcd.points)


def get_molecular_surface_point_cloud_const_density(centers: np.ndarray,
                                                    radii: Union[np.ndarray, List],
                                                    density: float = 0.3,
                                                    num_samples_per_atom: int = 25,
                                                    probe_radius: float = 1.2,
                                                    ball_radii: List[float] = [1.2],
                                                    color: List[float] = [1.0, 0.0, 0.0]
                                                    ) -> o3d.geometry.PointCloud:
    """
    Gets the point cloud representation of a molecule's van der Waals surface. Takes into account
    the vdW radii of different atoms. Removes overlapping points within vdW radii of neighboring
    atoms.

    Parameters
    ----------
    centers : np.ndarray (N, 3)
        Cartesian coordinates of the atom centers of a molecule.

    radii : np.ndarray (N,)
        van der Waals radii of each atom (in Angstrom) in the same order as the centers parameter.

    density : float (default = 0.3)
        The density of points on the surface. The number of points is calculated from the solvent
        accessible surface area approximately computed by the surface area of the generated mesh.

    num_samples_per_atom : int, optional
        Number of points to sample from the surface of each atom. Default is 25.
        Note that this value is scaled by a given atom's relative vdW radius to a carbon
        and SQUARED. Typically choose a value between 15 and 35. For example, if set to
        20, a carbon atom would have 400 sampled points.

    probe_radius : float, optional
        The radius of a probe atom to act as a "solvent accessible surface".
        Default is 1.2 angstroms which is the radius of a Hydrogen atom.

    ball_radii : list, optional
        The radius of the ball(s) used in Open3D's ball pivoting algorithm to generate
        a triangle mesh. Default is [1.2].

    color : list, optional
        RGB color values for the point cloud (default is [1., 0., 0.] red).

    Returns
    -------
    o3d.geometry.PointCloud
        Point cloud object representation of the molecular surface.
    """
    mesh, _ = _get_molecular_surface_mesh(centers=centers,
                                          radii=radii,
                                          num_samples_per_atom=num_samples_per_atom,
                                          probe_radius=probe_radius,
                                          ball_radii=ball_radii,
                                          color=color)
    # solv_acc_surf_area = rdFreeSASA.CalcSASA(mol, radii) # solvent accessible surface area
    solv_acc_surf_area = mesh.get_surface_area() # Approximate solvent accessible surface area
    num_points = int(density * solv_acc_surf_area)
    pcd = mesh.sample_points_poisson_disk(num_points)
    return pcd


def get_molecular_surface_const_density(centers:np.ndarray,
                                        radii:Union[np.ndarray, List],
                                        density: float = 0.3,
                                        num_samples_per_atom:int = 25,
                                        probe_radius: float = 1.2,
                                        ball_radii: List[float] = [1.2],
                                        ) -> np.ndarray:
    """
    Gets the point cloud representation of a molecule's van der Waals surface and outputs a
    numpy array. Takes into account the vdW radii of different atoms. Removes overlapping points
    within vdW radii of neighboring atoms.

    Parameters
    ----------
    centers : np.ndarray (N, 3)
        Cartesian coordinates of the atom centers of a molecule.

    radii : np.ndarray (N,)
        van der Waals radii of each atom (in Angstrom) in the same order as the centers parameter.

    density : float (default = 0.3)
        The density of points on the surface. The number of points is calculated from the solvent
        accessible surface area approximately computed by the surface area of the generated mesh.

    num_samples_per_atom : int, optional
        Number of points to sample from the surface of each atom. Default is 25.
        Note that this value is scaled by a given atom's relative vdW radius to a carbon
        and SQUARED. Typically choose a value between 15 and 35. For example, if set to
        20, a carbon atom would have 400 sampled points.

    probe_radius : float, optional
        The radius of a probe atom to act as a "solvent accessible surface".
        Default is 1.2 angstroms which is the radius of a Hydrogen atom.

    ball_radii : list, optional
        The radius of the ball(s) used in Open3D's ball pivoting algorithm to generate
        a triangle mesh. Default is [1.2].

    Returns
    -------
    np.ndarray
        Coordinates of points representing the molecular surface.
    """
    pcd = get_molecular_surface_point_cloud_const_density(centers=centers,
                                                          radii=radii,
                                                          density=density,
                                                          num_samples_per_atom=num_samples_per_atom,
                                                          probe_radius=probe_radius,
                                                          ball_radii=ball_radii)
    return np.asarray(pcd.points)


def get_electrostatics(mol: Chem.Mol, points: np.ndarray) -> np.ndarray:
    """
    Compute the Coulomb potential values at each point for a given molecule.
    Assumes the input "mol" already has an optimized conformer. Gets partial charges from
    MMFF or Gasteiger.

    Parameters
    ----------
    mol : rdkit.Chem.Mol object
        RDKit molecule object with an optimized geometry in conformers.

    points : np.ndarray (N, 3)
        Coordinates of sampled points to compute Coulomb potential at.

    Returns
    -------
    np.ndarray (N)
        Electrostatic potential values corresponding to each point.
    """
    try:
        mol.GetConformer()
    except ValueError as e:
        raise ValueError("Provided rdkit.Chem.Mol object did not have conformer embedded.", e)

    molec_props = Chem.AllChem.MMFFGetMoleculeProperties(mol)
    if molec_props:
        charges = np.array([molec_props.GetMMFFPartialCharge(i) for i, _ in enumerate(mol.GetAtoms())])
    else:
        print("MMFF charges not available for the input molecule, defaulting to Gasteiger charges.")
        AllChem.ComputeGasteigerCharges(mol)
        charges=np.array([a.GetDoubleProp('_GasteigerCharge') for a in mol.GetAtoms()])

    centers = mol.GetConformer().GetPositions()

    distances = np.linalg.norm(points[:, np.newaxis] - centers, axis=2)
    # Calculate the potentials
    E_pot = np.dot(charges, 1 / distances.T) * COULOMB_SCALING
    # Ensure that invalid distances (where distance is 0) are handled
    E_pot[np.isinf(E_pot)] = 0
    return E_pot


def get_electrostatics_given_point_charges(charges: np.ndarray,
                                           positions: np.ndarray,
                                           points: np.ndarray)-> np.ndarray:
    """
    Compute the Coulomb potential values at each point for a given set of charges at defined positions.

    Parameters
    ----------
    charges : np.ndarray (N,)
        Charges, with units [V]

    positions : np.ndarray (N, 3)
        Coordinates of point charges, with units of A.

    points : np.ndarray (M, 3)
        Coordinates of point cloud at which to compute electrostatic potential, with units of A.

    Returns
    -------
    np.ndarray (M,)
        Electrostatic potential values corresponding to each point.
    """

    distances = np.linalg.norm(points[:, np.newaxis] - positions, axis=2)
    # Calculate the potentials
    E_pot = np.dot(charges, 1. / distances.T) * COULOMB_SCALING # [eV/e] = [V]
    # Ensure that invalid distances (where distance is 0) are handled
    E_pot[np.isinf(E_pot)] = 0
    return E_pot


def color_pcd_with_electrostatics(pcd: o3d.geometry.PointCloud, E_pot: np.ndarray) -> None:
    """
    Color the point cloud based on elecrostatic potential. Colors are only scaled to the molecule
    itself (i.e., the colors are not comparable to different molecules).
    Red is positive, blue is negative, black is neutral.
    """

    colors = np.zeros((len(E_pot), 3))
    colors[:,0] = np.where(E_pot < 0, 0, E_pot/np.max(E_pot)).squeeze()
    colors[:,2] = np.where(E_pot >= 0, 0, -E_pot/np.max(-E_pot)).squeeze()
    pcd.colors = o3d.utility.Vector3dVector(colors)


def get_sample_atom_volume(R: float, num_samples: int) -> np.ndarray:
    """
    Sample points uniformly from a sphere.
    https://stackoverflow.com/questions/5408276/sampling-uniformly-distributed-random-points-inside-a-spherical-volume

    Parameters
    ----------
    R : float
        Radius
    num_samples : int
        Number of samples

    Returns
    -------
    np.ndarray (num_samples, 3)
    """
    phi = np.random.uniform(0, 2*np.pi, num_samples)
    cos_theta = np.random.uniform(-1, 1, num_samples)
    u = np.random.uniform(0,1, num_samples)

    theta = np.arccos(cos_theta)
    r = R * u**(1/3)
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)
    return np.column_stack((x, y, z))


def get_molecular_volume(centers: np.ndarray, radii: np.ndarray, num_samples: int = None, num_samples_per_atom: int=15):
    """
    Sample points uniformly from a molecule's volume. Does not include probe radius.
    Ignores possible overlap of vdW radius of atoms when randomly sub-selecting points for the
    volume.

    Parameters
    ----------
    centers : np.ndarray (N, 3)
        Cartesian coordinates of the atom centers of a molecule.

    radii : np.ndarray (N,)
        van der Waals radii of each atom (in Angstrom) in the same order as the centers parameter.

    num_samples_per_atom : int, optional
        Number of points to sample from the surface of each atom. Default is 15.
        Note that this value is scaled by a given atom's relative vdW radius to a carbon
        and CUBED. Typically choose a value between 10 and 20. For example, if set to
        20, a carbon atom would have 8000 sampled points.

    Returns
    -------
    np.ndarray
        Coordinates of points representing the molecular surface.
    """
    if num_samples_per_atom > 20:
        raise ValueError('Do not set num_samples_per_atom to be larger than 20 for performance\
                         issues. The number is cubed internally.')
    # cutoff = 1.4
    # radii = cutoff * (radii / 1.7)
    num_samples_per_atom = np.ceil((num_samples_per_atom * (radii / 1.7))**3)

    points = np.vstack([get_sample_atom_volume(r, int(num_samples_per_atom[i])) for i, r in enumerate(radii)])
    centers = np.repeat(centers, [int(n) for n in num_samples_per_atom], axis=0)
    points += centers # translate atoms to their centers

    # Subselect points by masking
    if num_samples is not None:
        idx = np.arange(len(points))
        np.random.shuffle(idx)
        if num_samples > len(points):
            num_samples = len(points)
        points = points[idx[:num_samples]]
    return points



"""
Script to tautomerize and protonate molecules using `molscrub` package developed by ForliLab.

https://github.com/forlilab/molscrub
"""

from typing import List

from rdkit import Chem
from rdkit.rdBase import BlockLogs

from shepherd_score.protonation.protonate import neutralize_atoms, remove_bad_protomers
from molscrub import Scrub


def tautomerize_molscrub(
    smiles: str,
    pH: float = 7.4,
    neutralize: bool = True,
) -> List[str]:
    """
    Find all tautomers/protomers of a molecule using `molscrub` package.

    Parameters
    ----------
    smiles : str
        SMILES string of the molecule to tautomerize/protonate.
    pH : float (default: 7.4)
        pH value to use for the protonation.
    neutralize : bool (default: True)
        Whether to neutralize the molecule before scrubbing.
    chemaxon_license_path : str | None (default: None)
        Path to the chemaxon license file.
        If ``None``, the ``CHEMAXON_LICENSE_URL`` environment variable is used.

    Returns
    -------
    list[str]
        List of SMILES strings of the tautomers/protomers.
    """
    scrub = Scrub(
        ph_low=pH,
        ph_high=pH,
        skip_gen3d=True,
    )

    _log_blocker = BlockLogs()

    mol = Chem.MolFromSmiles(smiles)
    if neutralize:
        mol = neutralize_atoms(mol)
    tautomers = scrub(mol)

    del _log_blocker

    return remove_bad_protomers([Chem.MolToSmiles(tautomer) for tautomer in tautomers])



import os
import subprocess

# Conditional JAX import - if not available, JAX-specific tests will be skipped
def _configure_jax_platform():
    """
    Configure JAX platform based on GPU availability.
    Sets JAX to CPU-only mode if no GPU is detected to avoid initialization errors.

    Returns:
        bool: True if GPU detected, False otherwise
    """
    try:
        # Try to detect GPU via nvidia-smi
        result = subprocess.run(
            ['nvidia-smi'],
            capture_output=True,
            text=True,
            timeout=2  # Add timeout to avoid hanging
        )
        gpu_detected = result.returncode == 0
    except (FileNotFoundError, subprocess.SubprocessError, subprocess.TimeoutExpired):
        gpu_detected = False

    if not gpu_detected:
        # Force JAX to use CPU only if no GPU detected
        os.environ['JAX_PLATFORMS'] = 'cpu'

    return gpu_detected



"""
Script to protonate SMILES strings using OpenBabel, MolScrub, or ChemAxon.

Requires (dependent on the method chosen):
- openbabel
- chemaxon
- molscrub
"""

import subprocess
import os
from typing import List, Literal
from rdkit import Chem
from copy import deepcopy

from rdkit.Chem.MolStandardize.rdMolStandardize import ChargeParent
from rdkit.rdBase import BlockLogs

# From ZINC22 build pipeline
BAD_CHARGES = [
    '[C-]',
    '[CH-]',
    '[CH2-]',
    '[C+]',
    '[CH+]',
    '[CH2+]',
    '[NH-]',
    '[c-]',
    '[cH-]',
    '[o+]',
    '[OH+]',
]


def neutralize_atoms(mol: Chem.Mol) -> Chem.Mol:
    """
    Attempts to neutralize every atom of a molecule by adding/removing hydrogens.
    It doesn't attempt to keep the formal charge of the original molecule.
    Neutralizes more than the rdMolStandardize.Uncharger.

    Below is copied from:
    https://rdkit.org/docs/Cookbook.html

    This neutralize_atoms() algorithm is adapted from Noel O'Boyle's nocharge code.
    It is a neutralization by atom approach and neutralizes atoms with a +1 or -1
    charge by removing or adding hydrogen where possible. The SMARTS pattern checks
    for a hydrogen in +1 charged atoms and checks for no neighbors with a negative
    charge (for +1 atoms) and no neighbors with a positive charge (for -1 atoms),
    this is to avoid altering molecules with charge separation (e.g., nitro groups).

    The neutralize_atoms() function differs from the `rdMolStandardize.Uncharger` behavior.
    See the MolVS documentation for Uncharger:

    https://molvs.readthedocs.io/en/latest/api.html#molvs-charge

    “This class uncharges molecules by adding and/or removing hydrogens.
    In cases where there is a positive charge that is not neutralizable,
    any corresponding negative charge is also preserved.”

    As an example, rdMolStandardize.Uncharger will not change charges on C[N+](C)(C)CCC([O-])=O,
    as there is a positive charge that is not neutralizable. In contrast, the neutralize_atoms()
    function will attempt to neutralize any atoms it can (in this case to C[N+](C)(C)CCC(=O)O).
    That is, neutralize_atoms() ignores the overall charge on the molecule, and attempts to
    neutralize charges even if the neutralization introduces an overall formal charge on the
    molecule.
    """
    mol = deepcopy(mol)
    pattern = Chem.MolFromSmarts("[+1!h0!$([*]~[-1,-2,-3,-4]),-1!$([*]~[+1,+2,+3,+4])]")
    at_matches = mol.GetSubstructMatches(pattern)
    at_matches_list = [y[0] for y in at_matches]
    if len(at_matches_list) > 0:
        for at_idx in at_matches_list:
            atom = mol.GetAtomWithIdx(at_idx)
            chg = atom.GetFormalCharge()
            hcount = atom.GetTotalNumHs()
            atom.SetFormalCharge(0)
            atom.SetNumExplicitHs(hcount - chg)
            atom.UpdatePropertyCache()
    return mol


def force_neutralize(mol: Chem.Mol) -> Chem.Mol:
    """
    Force neutralize a molecule by adding/removing hydrogens.
    Does not attempt to keep the formal charge of the original molecule.
    First runs `neutralize_atoms`, then runs `rdMolStandardize.ChargeParent`.
    """
    with BlockLogs():
        mol = deepcopy(mol)
        mol = neutralize_atoms(mol)
        mol = ChargeParent(mol)
    return mol


def remove_bad_protomers(protomers: List[str]) -> List[str]:
    """
    Remove protomers with bad charges from the rules in ZINC22 build pipeline.

    Parameters
    ----------
    protomers : list[str]
        List of SMILES strings of the protomers.

    Returns
    -------
    list[str]
        List of SMILES strings of the protomers with bad charges removed.
    """
    bad_charges = [c for c in BAD_CHARGES if c != '']
    passing_prots = []
    for smi in protomers:
        has_bad_charge = False
        for c in bad_charges:
            if c in smi:
                has_bad_charge = True
                break
        if not has_bad_charge:
            passing_prots.append(smi)
    return passing_prots


def protonate_smiles(smiles: str,
                     pH: float = 7.4,
                     method: Literal['openbabel', 'molscrub', 'chemaxon'] = 'molscrub',
                     *,
                     path_to_bin: str = '',
                     cxcalc_exe: str | None = None,
                     molconvert_exe: str | None = None,
                     chemaxon_license_path: str | None = None,
                     ) -> List[str]:
    """
    Protonate SMILES string with MolScrub, OpenBabel, or ChemAxon at given pH.

    ChemAxon requires `cxcalc` and `molconvert` executables to be installed with relevant license
    as input or set to the `CHEMAXON_LICENSE_URL` environment variable.

    OpenBabel workflow adapted from DockString:
    https://github.com/dockstring/dockstring/blob/main/dockstring/utils.py#L330

    Parameters
    ----------
    smiles : str
        SMILES string of molecule to be protonated.
    pH : float (default: 7.4)
        pH at which the molecule should be protonated.
    method : Literal['openbabel', 'molscrub', 'chemaxon']
        Method to use for protonation. Defaults to 'molscrub'.
    path_to_bin : str (default: '')
        Path to environment bin containing `mk_prepare_ligand.py`.
    cxcalc_exe : str | None (default: None)
        Path to cxcalc executable.
    molconvert_exe : str | None (default: None)
        Path to molconvert executable.
    chemaxon_license_path : str | None (default: None)
        Path to chemaxon license file.
        If ``None``, the ``CHEMAXON_LICENSE_URL`` environment variable is used.

    Returns
    -------
    list[str]
        List of SMILES strings of tautomers/protomers.
    """
    if method not in ['openbabel', 'molscrub', 'chemaxon']:
        raise ValueError(f'Invalid method: {method}')

    if method == 'openbabel':
        # cmd list format raises errors, therefore one string
        cmd = f'{path_to_bin}obabel -:"{smiles}" -ismi -ocan -p {pH}'
        cmd_return = subprocess.run(cmd, capture_output=True, shell=True)
        output = cmd_return.stdout.decode('utf-8')

        if cmd_return.returncode != 0:
            raise ValueError(f'Could not protonate SMILES: {smiles}')

        return [output.strip()]

    elif method == 'molscrub':
        from shepherd_score.protonation.molscrub_utils import tautomerize_molscrub
        return tautomerize_molscrub(smiles=smiles, pH=pH)

    elif method == 'chemaxon':
        if cxcalc_exe is None or molconvert_exe is None:
            raise ValueError('cxcalc_exe and molconvert_exe are required for chemaxon protonation')

        if os.environ.get('CHEMAXON_LICENSE_URL') is None or chemaxon_license_path is not None:
            raise ValueError('CHEMAXON_LICENSE_URL is not set')

        from shepherd_score.protonation.chemaxon_utils import tautomerize_chemaxon
        return tautomerize_chemaxon(
            smiles=smiles,
            pH=pH,
            cxcalc_exe=cxcalc_exe,
            molconvert_exe=molconvert_exe,
            chemaxon_license_path=chemaxon_license_path,
        )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
from datetime import datetime

# Add the project root to the path so autodoc can find the package
sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'shepherd-score'
copyright = f'2024-{datetime.now().year}, Kento Abeywardane'
author = 'Kento Abeywardane'

# The version info
from shepherd_score import __version__  # noqa: E402
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosummary',
    'sphinx.ext.mathjax',
    'sphinx_copybutton',
    'myst_nb',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']

# The master toctree document
master_doc = 'index'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_book_theme'
pygments_style = 'sphinx'

# Set HTML title without version number
html_title = 'shepherd-score'
html_logo = '_static/logo.svg'

html_theme_options = {
    'repository_url': 'https://github.com/coleygroup/shepherd-score',
    'path_to_docs': 'docs',
    'use_source_button': True,
    'use_download_button': True,
    'use_repository_button': True,
    'use_issues_button': True,
    'logo': {
        'image_light': '_static/logo.svg',
        'image_dark': '_static/logo.svg',
        'text': 'shepherd-score',
    },
    'icon_links': [
        {
            'name': 'GitHub',
            'url': 'https://github.com/coleygroup/shepherd-score',
            'icon': 'fa-brands fa-square-github',
            'type': 'fontawesome',
        },
        {
            'name': 'PyPI',
            'url': 'https://pypi.org/project/shepherd-score/',
            'icon': 'fa-solid fa-box',
            'type': 'fontawesome',
        },
    ],
}

html_static_path = ['_static']
html_css_files = ['custom.css']

# copybutton settings
copybutton_exclude = '.linenos, .gp'

# -- Extension configuration -------------------------------------------------

# Napoleon settings for NumPy-style docstrings
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# Autodoc settings
autodoc_default_options = {
    'members': True,
    'undoc-members': True,
    'show-inheritance': True,
    'member-order': 'bysource',
}
autodoc_typehints = 'description'
autodoc_mock_imports = [
    'torch',
    'jax',
    'jaxlib',
    'optax',
    'open3d',
    'rdkit',
    'meeko',
    'vina',
    'openbabel',
    'prolif',
    'biopython',
    'Bio',
    'molscrub',
    'py3Dmol',
    'sklearn',
]

# Intersphinx mapping
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'pandas': ('https://pandas.pydata.org/docs/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    'rdkit': ('https://www.rdkit.org/docs/', None),
}

# myst-nb settings
nb_execution_mode = 'off'  # Don't execute notebooks during build



"""
Functions used for SE(3) transformations. (Torch implementation).
Has support for operations with batches.

Namely, converting quaternions to rotation matrices, getting an SE(3) transform from SE(3)
parameters, and applying the SE(3) transformation on a set of points.

Credit to Lewis J. Martin as this was adapted from
https://github.com/ljmartin/align/blob/main/0.2%20aligning%20principal%20moments%20of%20inertia.ipynb
and PyTorch's implementations.
"""
import torch
import torch.nn.functional as F


def quaternions_to_rotation_matrix(quaternions: torch.Tensor) -> torch.Tensor:
    """
    Converts quaternion to a rotation matrix. Supports batched and non-batched inputs.
    Adapted from PyTorch3D:
    https://pytorch3d.readthedocs.io/en/latest/_modules/pytorch3d/transforms/rotation_conversions.html#quaternion_to_matrix

    Parameters
    ----------
    quaternions : torch.Tensor (batch, 4) or (4,)
        Quaternion parameters in (r,i,j,k) order. Accepts single set of parameters or a batched
        set.

    Returns
    -------
    rotation_matrix : torch.Tensor (batch, 3, 3) or (3,3)
        Rotation matrix converted from quaternion in batched or single instance form.
    """
    if quaternions.shape[-1] != 4:
        raise ValueError(f'Last dimension of "quaternions" must be length 4. Instead the shape given was: {quaternions.shape}')

    # Single instance
    if len(quaternions.shape) == 1:
        r, i, j, k = torch.unbind(quaternions, -1)
        two_s = torch.Tensor([2.0]).to(quaternions.device) / torch.sum(quaternions **2)

        rotation_matrix = torch.stack(
            (
                1 - two_s * (j * j + k * k),
                two_s * (i * j - k * r),
                two_s * (i * k + j * r),
                two_s * (i * j + k * r),
                1 - two_s * (i * i + k * k),
                two_s * (j * k - i * r),
                two_s * (i * k - j * r),
                two_s * (j * k + i * r),
                1 - two_s * (i * i + j * j),
            ),
            -1,
        )
        return rotation_matrix.reshape(quaternions.shape[:-1] + (3, 3))

    # Batched
    elif len(quaternions.shape) == 2:
        r, i, j, k = torch.unbind(quaternions, 1)
        two_s = torch.Tensor([2.0]).to(quaternions.device) / torch.sum(quaternions ** 2, dim=1)
        rotation_matrix = torch.stack(
            (
                1 - two_s * (j * j + k * k),
                two_s * (i * j - k * r),
                two_s * (i * k + j * r),
                two_s * (i * j + k * r),
                1 - two_s * (i * i + k * k),
                two_s * (j * k - i * r),
                two_s * (i * k - j * r),
                two_s * (j * k + i * r),
                1 - two_s * (i * i + j * j),
            ),
            1
        ).reshape((-1, 3, 3))
        return rotation_matrix
    else:
        raise ValueError(f'Input "quaternions" must be a 1D Tensor of length 4 or a batched version of shape (batch_size,4). Instead the shape given was: {quaternions.shape}')


def get_SE3_transform(se3_params: torch.Tensor
                     ) -> torch.Tensor:
    """ Constructs an SE(3) transformtion matrix from parameters.
    Supports batched and non-batched inputs.

    Parameters
    ----------
    se3_params : torch.Tensor (batch, 7) or (7,)
        Parameters for SE(3) transformation.
        The first 4 values in the last dimension are quaternions of form (r,i,j,k)
        and the last 3 values of the last dimension are the translations in (x,y,z).

    Returns
    -------
    se3_matrix : torch.Tensor (batch, 4, 4) or (4, 4)
        se3_params converted to a 4x4 SE(3) transformation matrix.
    """
    if se3_params.shape[-1] != 7:
        raise ValueError(f'Last dimension of "se3_params" must be length 7. Instead the shape given was: {se3_params.shape}')

    # Single instance
    if len(se3_params.shape) == 1:
        # Extract quaternion and translation parameters
        quaternion_params = se3_params[:4]
        translation_params = se3_params[4:]

        # Normalize quaternion to ensure unit length
        quaternion_params = F.normalize(quaternion_params, p=2, dim=-1)
        rotation_matrix = quaternions_to_rotation_matrix(quaternion_params)

        # Construct SE(3) transformation matrix
        se3_matrix = torch.eye(4, dtype=torch.float32, device=se3_params.device)
        se3_matrix[:3, :3] = rotation_matrix
        se3_matrix[:3, 3] = translation_params
        return se3_matrix

    # Batched
    elif len(se3_params.shape) == 2:
        # Extract quaternion and translation parameters
        quaternion_params = se3_params[:, :4]
        translation_params = se3_params[:, 4:]

        # Normalize quaternion to ensure unit length
        quaternion_params = F.normalize(quaternion_params, p=2, dim=1)
        rotation_matrix = quaternions_to_rotation_matrix(quaternion_params)

        # Construct SE(3) transformation matrix
        se3_matrix = torch.eye(4, device=se3_params.device).repeat((quaternion_params.shape[0],1,1))
        se3_matrix[:, :3, :3] = rotation_matrix
        se3_matrix[:, :3, 3] = translation_params
        return se3_matrix

    else:
        raise ValueError(f'Input "se3_params" must be a 1D Tensor of length 7 or a batched version of shape (batch_size,7). Instead the shape given was: {se3_params.shape}')


def apply_SE3_transform(points: torch.Tensor,
                        SE3_transform: torch.Tensor
                       ) -> torch.Tensor:
    """
    Takes a point cloud and transforms it according to the provided SE3 transformation matrix.
    Supports batched and non-batched inputs.

    Parameters
    ----------
    points : torch.Tensor (batch, N, 3) or (N, 3)
        Set of coordinates representing a point cloud.
    SE3_transform : torch.Tensor (batch, 4, 4) or (4, 4)
        SE(3) transformation matrix. If 'points' argument is batched, this one should be too.

    Returns
    -------
    transformed_points : torch.Tensor (batch, N, 3) or (N, 3)
        Set of coordinates transformed by the corresponding SE(3) transformation.
    """
    if points.shape[-1] != 3:
        raise ValueError(f'"points" should have shape (N_points, 3) or (batch, N_points, 3). Instead the shape given was: {points.shape}')
    if SE3_transform.shape[-2:] != (4,4):
        raise ValueError(f'"SE3_transform" should have shape (4, 4) or (batch, 4, 4). Instead the shape given was: {SE3_transform.shape}')
    if len(SE3_transform.shape) != len(points.shape):
        raise ValueError(f'Shapes of points and SE3_transform should be the same length. Instead {len(SE3_transform.shape)} and {len(points.shape)} were given.')

    # Single instance
    if len(SE3_transform.shape) == 2:
        transformed_points = (SE3_transform[:3,:3] @ points.T).T + SE3_transform[:3,3]
        return transformed_points

    # Batched
    elif len(SE3_transform.shape):
        rotated_points = torch.bmm(SE3_transform[:, :3,:3], points.permute(0,2,1)).permute(0,2,1)
        transformed_points = rotated_points + SE3_transform[:, None, :3, 3]
        return transformed_points

    else:
        raise ValueError(f'"points" and "SE3_transform" must be either batched or a single instance. \
        The expected length of shape for both should be 2 (single instance or 3 (batch) but {len(SE3_transform)} was given.')


def apply_SO3_transform(points: torch.Tensor,
                        SE3_transform: torch.Tensor
                        ) -> torch.Tensor:
    """
    Takes a point cloud and ONLY ROTATES it according to the provided SE3 transformation matrix.
    Supports batched and non-batched inputs.

    Parameters
    ----------
    points : torch.Tensor (batch, N, 3) or (N, 3)
        Set of coordinates representing a point cloud.
    SE3_transform : torch.Tensor (batch, 4, 4) or (4, 4)
        SE(3) transformation matrix. If 'points' argument is batched, this one should be too.

    Returns
    -------
    rotated_points : torch.Tensor (batch, N, 3) or (N, 3)
        Set of coordinates rotated by the rotation component of the SE(3) transformation.
    """
    if points.shape[-1] != 3:
        raise ValueError(f'"points" should have shape (N_points, 3) or (batch, N_points, 3). Instead the shape given was: {points.shape}')
    if SE3_transform.shape[-2:] != (4,4):
        raise ValueError(f'"SE3_transform" should have shape (4, 4) or (batch, 4, 4). Instead the shape given was: {SE3_transform.shape}')
    if len(SE3_transform.shape) != len(points.shape):
        raise ValueError(f'Shapes of points and SE3_transform should be the same length. Instead {len(SE3_transform.shape)} and {len(points.shape)} were given.')

    # Single instance
    if len(SE3_transform.shape) == 2:
        rotated_points = (SE3_transform[:3,:3] @ points.T).T
        return rotated_points

    # Batched
    elif len(SE3_transform.shape):
        rotated_points = torch.bmm(SE3_transform[:, :3,:3], points.permute(0,2,1)).permute(0,2,1)
        return rotated_points

    else:
        raise ValueError(f'"points" and "SE3_transform" must be either batched or a single instance. \
        The expected length of shape for both should be 2 (single instance or 3 (batch) but {len(SE3_transform)} was given.')



"""
Autodock Vina Docking evaluation pipeline.

Adapted from Therapeutic Data Commons (TDC).
Huang et al. (2021) https://arxiv.org/abs/2102.09548

Requires:
- vina
- meeko
- openbabel (if protonating ligands)
"""
import os
import time
from typing import Tuple, Optional, Literal, List
from pathlib import Path
import uuid

import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem

from shepherd_score.conformer_generation import update_mol_coordinates
from shepherd_score.protonation.protonate import protonate_smiles

try:
    from vina import Vina
except ImportError:
    raise ImportError(
        "Please install vina following guidance in https://github.com/ccsb-scripps/AutoDock-Vina/tree/develop/build/python"
    )

try:
    from meeko import MoleculePreparation
    from meeko import PDBQTWriterLegacy
    from meeko import PDBQTMolecule
    from meeko import RDKitMolCreate

except ImportError:
    raise ImportError(
        "Please install meeko following guidance in https://meeko.readthedocs.io/en/release-doc/installation.html"
    )

def embed_conformer_from_smiles_fixed(
    smiles: str,
    attempts: int=50,
    MMFF_optimize: bool=True,
    random_seed: int=123456789,
) -> Chem.Mol:
    """
    Embeds a mol object into a 3D RDKit mol object with ETKDG (and optional MMFF94)
    """
    mol = Chem.MolFromSmiles(smiles)
    mol = Chem.AddHs(mol)
    AllChem.EmbedMolecule(mol, randomSeed=random_seed, maxAttempts=attempts)
    if MMFF_optimize:
        AllChem.MMFFOptimizeMolecule(mol, maxIters=200)
    return mol

class VinaBase:
    """
    Base class for Vina scoring function.
    """
    def __init__(
        self,
        receptor_pdbqt_file: str,
        center: Tuple[float],
        box_size: Tuple[float],
        pH: float = 7.4,
        scorefunction: str = "vina",
        num_processes: int = 4,
        verbose: int = 0,
        *,
        protonate_method: Literal['openbabel', 'molscrub', 'chemaxon'] = 'molscrub',
        path_to_bin: str = '',
        cxcalc_exe: str | None = None,
        molconvert_exe: str | None = None,
        chemaxon_license_path: str | None = None,
    ):
        """
        Constructs Vina scoring function with receptor.

        Arguments
        ---------
        receptor_pdbqt_file : str path to .pdbqt file of receptor.
        center : Tuple[float] (len=3) coordinates for the center of the pocket.
        box_size : Tuple[float](len=3) box edge lengths of pocket.
        pH : float Experimental pH used for crystal structure elucidation.
        scorefunction : str (default=vina) name of scoring function to use with Vina. 'vina' or 'ad4'
        num_processes : int (default=2) Number of cpus to use for scoring
        verbose : int (default = 0) Level of verbosity from vina.Vina (0 is silent)

        protonate_method : Literal['openbabel', 'molscrub', 'chemaxon'] (default = 'molscrub') method to use for protonation
        cxcalc_exe : str | None (default = None) path to cxcalc executable
        molconvert_exe : str | None (default = None) path to molconvert executable
        chemaxon_license_path : str | None (default = None) path to chemaxon license file
        """
        self.v = Vina(sf_name=scorefunction, seed=987654321, verbosity=verbose, cpu=num_processes)
        self.receptor_pdbqt_file = receptor_pdbqt_file
        self.center = center
        self.box_size = box_size
        self.pH = pH
        self.v.set_receptor(rigid_pdbqt_filename=receptor_pdbqt_file)
        try:
            self.v.compute_vina_maps(center=self.center, box_size=self.box_size)
        except Exception:
            raise ValueError(
                "Cannot compute the affinity map, please check center and box_size"
            )
        self.mk_prep_ligand = MoleculePreparation()

        self.protonate_method = protonate_method
        self.path_to_bin = path_to_bin
        self.cxcalc_exe = cxcalc_exe
        self.molconvert_exe = molconvert_exe
        self.chemaxon_license_path = chemaxon_license_path

        self.state = None

    def load_ligand_from_smiles(
        self,
        ligand_smiles: str,
        protonate: bool = False,
        return_all: bool = False,
    ) -> List[Chem.Mol]:
        """
        Load ligand SMILES string into Vina.
        """
        if protonate:
            protomers = protonate_smiles(
                smiles=ligand_smiles,
                pH=self.pH,
                method=self.protonate_method,
                path_to_bin=self.path_to_bin,
                cxcalc_exe=self.cxcalc_exe,
                molconvert_exe=self.molconvert_exe,
                chemaxon_license_path=self.chemaxon_license_path,
            )
            if return_all:
                ligand_smiles = protomers
            else:
                ligand_smiles = [protomers[0]]
        else:
            ligand_smiles = [ligand_smiles]

        mols = []
        for smi in ligand_smiles:
            m = embed_conformer_from_smiles_fixed(smi, MMFF_optimize=True, random_seed=123456789)
            if m is not None:
                mols.append(m)
        return mols

    def load_ligand_from_sdf(
        self,
        sdf_file: str,
    ) -> Chem.Mol:
        """
        Load ligand from SDF file into Vina.
        """
        mol = Chem.SDMolSupplier(sdf_file, removeHs=False)[0]

        if mol.GetNumConformers() == 0:
            mols = self.load_ligand_from_smiles(Chem.MolToSmiles(mol))
            if len(mols) > 0:
                mol = mols[0]
            else:
                raise ValueError(
                    f"Failed to load SDF file and could not embed conformer: {sdf_file}"
                )
        return mol

    def _prep_ligand(
        self,
        ligand: Chem.Mol,
    ) -> str | None:
        """
        Prepare ligand for docking.
        """
        try:
            molsetup = self.mk_prep_ligand.prepare(ligand)[0]
            ligand_pdbqt_string, was_successful, error_message = PDBQTWriterLegacy.write_string(molsetup)
            if not was_successful:
                print(error_message)
                return None
            return ligand_pdbqt_string
        except Exception as e:
            print(e)
            return None

    def _center_ligand(
        self,
        ligand: Chem.Mol,
        center: Tuple[float, float, float],
    ) -> Chem.Mol:
        """
        Centers a ligand conformer's center of mass to a given center.
        Returns a centered copy of the ligand.
        """
        ligand_com = ligand.GetConformer().GetPositions().mean(axis=0)
        recentered_coords = ligand.GetConformer().GetPositions() - ligand_com + center
        ligand = update_mol_coordinates(ligand, recentered_coords)
        return ligand

    def dock_ligand(
        self,
        ligand: Chem.Mol,
        output_file: Optional[str] = None,
        exhaustiveness: int = 8,
        n_poses: int = 5,
    ) -> Tuple[np.float64, np.float64, Chem.Mol] | None:
        """
        Given a ligand, do a global optimization and return the best energy and optionally the pose.

        Arguments
        ---------
        ligand : Chem.Mol ligand to dock.
        output_file : Optional[str] path to save docked poses.
        exhaustiveness : int (default = 8) Number of Monte Carlo simulations to run per pose.
        n_poses : int (default = 5) Number of poses to save.

        Returns
        -------
        Tuple
            total_energy : np.float64 energy of the best pose (kcal/mol)
            torsion_energy : np.float64 torsion energy of the best pose (kcal/mol)
            docked_rdmol : Chem.Mol rdkit mol of best docked pose
        """
        try:
            ligand_pdbqt_string = self._prep_ligand(ligand)
            self.v.set_ligand_from_string(ligand_pdbqt_string)
            self.v.dock(exhaustiveness=exhaustiveness, n_poses=n_poses)
            self.state = 'docked'
            if output_file is not None:
                self.v.write_poses(str(output_file), n_poses=n_poses, overwrite=True)

            # Extract docked poses to rdkit mols
            vina_output_string = self.v.poses()
            docked_pdbqt_mols = PDBQTMolecule(vina_output_string, is_dlg=False, skip_typing=False)
            docked_rdmol = RDKitMolCreate.from_pdbqt_mol(docked_pdbqt_mols)[0]

        except Exception as e:
            print(e)
            return np.nan, np.nan, None
        (total_energy, _, _, _, _, _, torsion_energy, _) = self.v.score()
        return total_energy, torsion_energy, docked_rdmol

    def score_ligand(
        self,
        ligand: Chem.Mol,
        center: bool | Tuple[float, float, float] = False,
    ) -> Tuple[np.float64, np.float64]:
        """
        Scores a given ligand's pose in it's current conformation (e.g., no optimization).

        Arguments
        ---------
        ligand : Chem.Mol ligand to score.
        center : bool | Tuple[float, float, float] | None (default = False)
            If a tuple, centers to those coordinates.
            If True, centers the ligand to the receptor's center.
            If False, does not translate the ligand from its initial conformation.

        Returns
        -------
        (total_energy, torsion_energy) : Tuple[np.float64, np.float64]
            Energies in kcal/mol
        """
        if center is True:
            ligand = self._center_ligand(ligand, self.center)
        elif isinstance(center, tuple):
            ligand = self._center_ligand(ligand, center)

        try:
            ligand_pdbqt_string = self._prep_ligand(ligand)
            self.v.set_ligand_from_string(ligand_pdbqt_string)
            (total_energy, _, _, _, _, _, torsion_energy, _) = self.v.score()
            self.state = None
        except Exception as e:
            print(e)
            return np.nan, np.nan
        return total_energy, torsion_energy

    def optimize_ligand(
        self,
        ligand: Chem.Mol,
        center: bool | Tuple[float, float, float] = False,
        max_steps: int | None = 10000,
        output_file: Optional[str] = None,
    ) -> Tuple[np.float64, np.float64, Chem.Mol]:
        """
        Locally optimize loaded ligand pose.

        Arguments
        ---------
        ligand : Chem.Mol ligand to optimize.
        center : bool | Tuple[float, float, float] | None (default = False)
            If a tuple, centers to those coordinates.
            If True, centers the ligand to the receptor's center.
            If False, does not translate the ligand from its initial conformation.
        max_steps : int | None (default = 10000) Maximum number of steps to take in the optimization.
            If None, uses the default value of 10000.
        output_file : Optional[str] path to save optimized pose.

        Returns
        -------
        Tuple
            total_energy : np.float64 energy of the best pose (kcal/mol)
            torsion_energy : np.float64 torsion energy of the best pose (kcal/mol)
            optimized_rdmol : Chem.Mol rdkit mol of optimized pose
        """
        if center is True:
            ligand = self._center_ligand(ligand, self.center)
        elif isinstance(center, tuple):
            ligand = self._center_ligand(ligand, center)

        _used_temp_file = False

        try:
            ligand_pdbqt_string = self._prep_ligand(ligand)
            self.v.set_ligand_from_string(ligand_pdbqt_string)
            (total_energy, _, _, _, _, _, torsion_energy, _) = self.v.optimize(
                max_steps=max_steps if max_steps is not None else 0,
            )
            self.state = 'optimized'
            if output_file is not None:
                self.v.write_pose(output_file, overwrite=True)

            if output_file is None:
                _used_temp_file = True
                _file_name = str(uuid.uuid4()) + ''.join(str(time.time()).split('.')[1])
                _file_name = f'{_file_name}_optimized.pdbqt'
                if os.environ.get('TMPDIR', None) is not None:
                    _dir_path = os.environ['TMPDIR']
                elif os.environ.get('/tmp', None) is not None:
                    _dir_path = '/tmp'
                else:
                    _dir_path = './'
                temp_output_file = str(Path(_dir_path) / _file_name)
                self.v.write_pose(temp_output_file, overwrite=True)

            if output_file is None:
                output_file = temp_output_file

            pdbqt_mol_opt = PDBQTMolecule.from_file(output_file)
            rdkitmol_opt = RDKitMolCreate.from_pdbqt_mol(pdbqt_mol_opt)[0]

            if _used_temp_file:
                os.remove(output_file)

        except Exception as e:
            print(e)
            return np.nan, np.nan, None
        return total_energy, torsion_energy, rdkitmol_opt

    def save_pose_to_file(self, output_file: str, n_poses: int = 1):
        if self.state is None:
            print("Cannot save pose in state None. Run docking or optimization first.")
            return
        if self.state == 'docked':
            self.v.write_poses(output_file, n_poses=n_poses, overwrite=True)
        elif self.state == 'optimized':
            self.v.write_pose(output_file, overwrite=True)


class VinaSmiles(VinaBase):
    """
    Perform docking search from a SMILES string.

    Adapted from TDC.
    """
    def __init__(self,
                 receptor_pdbqt_file: str,
                 center: Tuple[float],
                 box_size: Tuple[float],
                 pH: float = 7.4,
                 scorefunction: str = "vina",
                 num_processes: int = 4,
                 verbose: int = 0,
                 *,
                 protonate_method: Literal['openbabel', 'molscrub', 'chemaxon'] = 'molscrub',
                 cxcalc_exe: str | None = None,
                 molconvert_exe: str | None = None,
                 chemaxon_license_path: str | None = None,
                 ):
        """
        Constructs Vina scoring function with receptor.

        Arguments
        ---------
        receptor_pdbqt_file : str path to .pdbqt file of receptor.
        center : Tuple[float] (len=3) coordinates for the center of the pocket.
        box_size : Tuple[float](len=3) box edge lengths of pocket.
        pH : float Experimental pH used for crystal structure elucidation.
        scorefunction : str (default=vina) name of scoring function to use with Vina. 'vina' or 'ad4'
        num_processes : int (default=2) Number of cpus to use for scoring
        verbose : int (default = 0) Level of verbosity from vina.Vina (0 is silent)

        protonate_method : Literal['openbabel', 'molscrub', 'chemaxon'] (default = 'molscrub') method to use for protonation
        cxcalc_exe : str | None (default = None) path to cxcalc executable
        molconvert_exe : str | None (default = None) path to molconvert executable
        chemaxon_license_path : str | None (default = None) path to chemaxon license file
        """
        super().__init__(
            receptor_pdbqt_file=receptor_pdbqt_file,
            center=center,
            box_size=box_size,
            pH=pH,
            scorefunction=scorefunction,
            num_processes=num_processes,
            verbose=verbose,
            protonate_method=protonate_method,
            cxcalc_exe=cxcalc_exe,
            molconvert_exe=molconvert_exe,
            chemaxon_license_path=chemaxon_license_path,
        )


    def __call__(self,
                 ligand_smiles: str,
                 output_file: Optional[str] = None,
                 exhaustiveness: int = 8,
                 n_poses: int = 5,
                 protonate: bool = False,
                 return_best_protomer: bool = False,
                 ) -> Tuple[float, Chem.Mol]:
        """
        Score ligand by docking in receptor.

        Arguments
        ---------
        ligand_smiles : str SMILES of ligand to dock.
        output_file : Optional[str] path to save docked poses.
        exhaustiveness : int (default = 8) Number of Monte Carlo simulations to run per pose.
        n_poses : int (default = 5) Number of poses to save.
        protonate : bool (default = False) (de-)protonate ligand with OpenBabel at pH=7.4
        return_best_protomer: bool (default = False) Evaluate all protomers and return the best
        energy and pose / protomer which may be different from the input SMILES.

        Returns
        -------
        Tuple
            float : energy (affinity) in kcal/mol
            Chem.Mol : docked ligand
        """
        if not return_best_protomer:
            ligand = self.load_ligand_from_smiles(ligand_smiles, protonate=protonate, return_all=False)
            total_energy, _, docked_mol = self.dock_ligand(
                ligand=ligand,
                output_file=output_file,
                exhaustiveness=exhaustiveness,
                n_poses=n_poses,
            )
            return total_energy, docked_mol
        else:
            protomers = self.load_ligand_from_smiles(ligand_smiles, protonate=protonate, return_all=True)
            best_energy = np.inf
            best_protomer = None
            for protomer in protomers:
                total_energy, _, docked_mol = self.dock_ligand(
                    ligand=protomer,
                    output_file=output_file,
                    exhaustiveness=exhaustiveness,
                    n_poses=n_poses,
                )
                if total_energy < best_energy:
                    best_energy = total_energy
                    best_protomer = docked_mol
            return best_energy, best_protomer



"""
Handles anything related to generating conformers with xTB or MMFF94.

Requires xtb installation with command-line access.
See https://xtb-docs.readthedocs.io/en/latest/setup.html for installation instructions.
"""
import os
from copy import deepcopy
import re
import shutil
import subprocess
import time
from pathlib import Path
from tqdm import tqdm
import uuid
from typing import Optional, List
import contextlib
import multiprocessing

from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Geometry import Point3D
from rdkit.Chem import rdMolAlign
from rdkit.ML.Cluster import Butina # type: ignore

TMPDIR = os.environ.get('TMPDIR', '')
if TMPDIR == '' and Path('/tmp').is_dir():
    TMPDIR = '/tmp'
TMPDIR = Path(TMPDIR)


@contextlib.contextmanager
def set_thread_limits(num_threads: int):
    """Temporarily set threading environment variables."""
    env_vars = [
        'OMP_NUM_THREADS',
        'OPENBLAS_NUM_THREADS',
        'MKL_NUM_THREADS',
        'NUMEXPR_NUM_THREADS',
        'VECLIB_MAXIMUM_THREADS',
    ]
    old_env = {var: os.environ.get(var) for var in env_vars}
    try:
        for var in env_vars:
            os.environ[var] = str(num_threads)
        yield
    finally:
        for var, val in old_env.items():
            if val is None:
                os.environ.pop(var, None)
            else:
                os.environ[var] = val


def update_mol_coordinates(mol: Chem.Mol, coordinates) -> Chem.Mol:
    """
    Update the coordinates of a 3D RDKit mol object with a new set of coordinates.

    Parameters
    ----------
    mol : Chem.Mol
        RDKit mol object with 3D coordinates to be replaced.
    coordinates : list or array-like
        List/array of new [x, y, z] coordinates.

    Returns
    -------
    Chem.Mol
        RDKit mol object with updated 3D coordinates.
    """
    mol_new = deepcopy(mol)
    conf = mol_new.GetConformer()
    for i in range(mol_new.GetNumAtoms()):
        x,y,z = coordinates[i]
        conf.SetAtomPosition(i, Point3D(x,y,z))
    return mol_new

def read_multi_xyz_file(file_dir: str):
    """
    Read an xyz file that potentially contains multiple structures.

    Parameters
    ----------
    file_dir : str
        Path to .xyz file.

    Returns
    -------
    all_coordinates : list
        List of lists containing the coordinates of each structure in the xyz file.
    all_elements : list
        List of lists containing the element types of each atom in each structure.
    """
    atom_types = [Chem.GetPeriodicTable().GetElementSymbol(i) for i in range(1, 119)]
    with open(file_dir, 'r') as file:
        lines = file.readlines()
        all_coordinates = []
        all_elements = []

        begin_coord = False
        coordinates = []
        elements = []
        for line in lines:
            stripped_line = (' '.join(line.split()))
            stripped_line_split = stripped_line.split(' ')
            numbers = re.findall(r"[-+]?(?:\d*\.\d+|\d+)", stripped_line)
            if len(stripped_line) == 0:
                continue

            if (len(numbers) == 3) & (stripped_line_split[0] in atom_types):
                begin_coord = True
                x,y,z = numbers
                coordinates.append([float(x),float(y),float(z)])
                elements.append(stripped_line_split[0])
            else:
                if begin_coord:
                    all_coordinates.append(coordinates)
                    all_elements.append(elements)
                    elements = []
                    coordinates = []
                begin_coord = False

    if begin_coord:
        all_coordinates.append(coordinates)
        all_elements.append(elements)

    return all_coordinates, all_elements


def embed_conformer(mol: Chem.Mol, attempts: int=50, MMFF_optimize: bool=False):
    """
    Embed a mol object into a 3D RDKit mol object with ETKDG (and optional MMFF94).

    Parameters
    ----------
    mol : Chem.Mol
        RDKit Mol object.
    attempts : int, optional
        Number of embedding attempts. Default is 50.
    MMFF_optimize : bool, optional
        Whether to optimize embedded conformer with MMFF94. Default is ``False``.

    Returns
    -------
    Chem.Mol or None
        RDKit mol object with 3D coordinates, or ``None`` if embedding fails.
    """
    try:
        mol = Chem.AddHs(mol)
        AllChem.EmbedMolecule(mol, maxAttempts = attempts)
        if MMFF_optimize:
            AllChem.MMFFOptimizeMolecule(mol)

        mol.GetConformer() # test whether conformer generation succeeded
    except Exception:
        return None

    return mol


def embed_conformer_from_smiles(smiles: str, attempts: int = 50, MMFF_optimize: bool = False):
    """
    Embed a SMILES into a 3D RDKit mol object with ETKDG (and optionally MMFF94).

    Parameters
    ----------
    smiles : str
        SMILES string of molecule.
    attempts : int, optional
        Number of embedding attempts. Default is 50.
    MMFF_optimize : bool, optional
        Whether to optimize embedded conformer with MMFF94. Default is ``False``.

    Returns
    -------
    Chem.Mol or None
        RDKit mol object with 3D coordinates, or ``None`` if embedding fails.
    """
    try:
        mol = Chem.MolFromSmiles(smiles)
    except Exception as e:
        print('Error in SMILES string when embedding molecule:', e)
        return None
    mol = embed_conformer(mol, attempts, MMFF_optimize)
    return mol


def conf_to_mol(mol, conf_id):
    """
    Convert a conformer of a RDKit mol object into its own RDKit mol object.

    Parameters
    ----------
    mol : Chem.Mol
        RDKit mol object with multiple conformers.
    conf_id : int
        ID of conformer to be converted into its own mol object.

    Returns
    -------
    Chem.Mol
        Mol object with only 1 conformer (the selected conformer).
    """
    conf = mol.GetConformer(conf_id)
    new_mol = Chem.Mol(mol)
    new_mol.RemoveAllConformers()
    new_mol.AddConformer(Chem.Conformer(conf))
    return new_mol


def generate_conformer_ensemble(mol_3d: Chem.Mol,
                                num_confs: int=100,
                                num_threads: int = 4,
                                threshold: float = 0.25,
                                num_opt_steps: int = 200):
    """
    Use ETKDG algorithm to embed multiple conformers from a given 3D conformer template.

    Optionally optimizes each embedded conformer with MMFF94.

    Parameters
    ----------
    mol_3d : Chem.Mol
        RDKit mol object with 3D coordinates.
    num_confs : int, optional
        Maximum number of conformers to be embedded with ETKDG. Default is 100.
    num_threads : int, optional
        Number of processors to be used in parallel when embedding conformers. Default is 4.
    threshold : float, optional
        RMSD threshold used to eliminate redundant conformers after ETKDG embedding.
        Default is 0.25.
    num_opt_steps : int, optional
        Number of MMFF94 optimization steps. Default is 200.

    Returns
    -------
    list
        List of mol objects, each containing 1 (unique) conformer.
    """
    mol_3d = deepcopy(mol_3d)

    cids = AllChem.EmbedMultipleConfs(
        mol_3d,
        clearConfs=True,
        numConfs=num_confs,
        pruneRmsThresh = threshold,
        maxAttempts = 50,
        numThreads = num_threads,
    )

    if num_opt_steps > 0:
        for cid in cids:
            AllChem.MMFFOptimizeMolecule(
                mol_3d,
                confId=cid,
                mmffVariant='MMFF94',
                maxIters = num_opt_steps,

            )
    mols = [conf_to_mol(mol_3d, c) for c in cids]

    return mols


def cluster_conformers_butina(conformers, threshold: float = 0.2, num_max_conformers: int = 100):
    """
    Cluster a list of conformers by their pairwise RMSD with Butina Clustering algorithm.

    Parameters
    ----------
    conformers : list
        List of rdkit mol objects containing conformers of a common molecule to be clustered.
    threshold : float, optional
        Initial RMSD threshold for clustering. Default is 0.2.
    num_max_conformers : int, optional
        Maximum number of conformers in the final clustered ensemble. Default is 100.

    Returns
    -------
    list
        List of int indices of the centroids of each cluster, to be indexed into conformers.
    """
    dists = []
    for i, conformer_i in enumerate(conformers):
        for j in range(i):
            mol_i = Chem.RemoveHs(deepcopy(conformer_i))
            mol_j = Chem.RemoveHs(deepcopy(conformers[j]))
            dists.append(rdMolAlign.GetBestRMS(mol_i, mol_j))

    clusts = Butina.ClusterData(dists, len(conformers), threshold, isDistData=True, reordering=True)

    if isinstance(num_max_conformers, int):
        while len(clusts) > num_max_conformers:
            threshold += 0.1
            clusts = Butina.ClusterData(dists, len(conformers),
                                        threshold, isDistData=True, reordering=True)

    select_confs = [c[0] for c in clusts] # picking centroids of each cluster

    return select_confs


def optimize_conformer_with_xtb(conformer: Chem.Mol,
                                solvent: Optional[str] = None,
                                num_cores: int = 1,
                                charge: int = 0,
                                temp_dir: str | Path = TMPDIR):
    """
    Use external calls to GFN2-XTB (command line) to optimize a conformer geometry.

    Parameters
    ----------
    conformer : Chem.Mol
        RDKit mol object containing 3D coordinates.
    solvent : str, optional
        Implicit solvent to be used during optimization. Must be a solvent supported
        by XTB (https://xtb-docs.readthedocs.io/en/latest/gbsa.html). Default is ``None``.
    num_cores : int, optional
        Number of CPU cores to be used in the xTB geometry optimization. Default is 1.
    charge : int, optional
        Molecular charge. Default is 0.
    temp_dir : str or Path, optional
        Temporary directory for I/O. Default is the system temporary directory.

    Returns
    -------
    tuple
        (xtb_mol, energy, charges) - tuple of optimized RDKit mol object,
        xTB energy (in Hartrees), and partial charges (in e-).
    """
    mol = deepcopy(conformer)

    with set_thread_limits(num_cores):
        # rand = str((os.getpid() * int(time.time())) % 123456789)
        rand = str(uuid.uuid4()) + ''.join(str(time.time()).split('.')[1])
        out_dir = Path(f'temp_xtb_opt_{rand}/')
        try:
            if temp_dir != '':
                if isinstance(temp_dir, str):
                    temp_dir = Path(temp_dir)
                out_dir = temp_dir / out_dir
            out_dir.mkdir(exist_ok=True)

            input_file = 'input_mol.xyz'
            Chem.rdmolfiles.MolToXYZFile(mol, str(out_dir/input_file))

            if solvent is not None:
                subprocess.check_call(
                    ['xtb', input_file, '--opt', '--alpb', solvent, '--parallel', str(num_cores), '--chrg', str(charge)],
                    cwd = out_dir,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.STDOUT,
                )
            else:
                subprocess.check_call(
                    ['xtb', input_file, '--opt', '--parallel', str(num_cores), '--chrg', str(charge)],
                    cwd = out_dir,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.STDOUT,
                )

            xtb_coords_list, xtb_elements_list = read_multi_xyz_file(out_dir/'xtbopt.xyz')
            xtb_coords = xtb_coords_list[0]
            xtb_mol = update_mol_coordinates(mol, xtb_coords)

            with open(out_dir/'xtbopt.xyz', 'r') as file:
                lines = file.readlines()
                for line in lines:
                    if 'energy' in line:
                        numbers = re.findall(r"[-+]?(?:\d*\.\d+|\d+)", line)
                        energy = float(numbers[0])
                        break

            with open(out_dir/'charges', 'r') as file:
                lines = file.readlines()
                charges = [0]*len(lines)
                for i, line in enumerate(lines):
                    charges[i] = float(line.split()[0])
                    xtb_mol.GetAtomWithIdx(i).SetProp('charge', str(charges[i]))

            xtb_mol.SetProp("energy", str(energy))

        finally:
            if out_dir.exists():
                shutil.rmtree(out_dir)

        return (xtb_mol, energy, charges)


def optimize_conformer_with_xtb_from_xyz_block(xyz_block: str,
                                               solvent: Optional[str] = None,
                                               num_cores: int = 1,
                                               charge: int = 0,
                                               temp_dir: str | Path = TMPDIR):
    """
    Use external calls to GFN2-XTB (command line) to optimize coordinates from an xyz block.

    Parameters
    ----------
    xyz_block : str
        String of an xyz block.
    solvent : str, optional
        Implicit solvent to be used during optimization. Must be a solvent supported
        by XTB (https://xtb-docs.readthedocs.io/en/latest/gbsa.html). Default is ``None``.
    num_cores : int, optional
        Number of CPU cores to be used in the xTB geometry optimization. Default is 1.
    charge : int, optional
        Molecular charge. Default is 0.
    temp_dir : str or Path, optional
        Temporary directory for I/O. Default is the system temporary directory.

    Returns
    -------
    tuple
        (xtb_xyz_block, energy, charges) - tuple of optimized xyz block string,
        xTB energy (in Hartrees), and partial charges (in e-).
    """
    with set_thread_limits(num_cores):
        # rand = str((os.getpid() * int(time.time())) % 123456789)
        rand = str(uuid.uuid4()) + ''.join(str(time.time()).split('.')[1])
        out_dir = Path(f'temp_xtb_opt_{rand}/')
        try:
            if temp_dir != '':
                if isinstance(temp_dir, str):
                    temp_dir = Path(temp_dir)
                out_dir = temp_dir / out_dir
            out_dir.mkdir(exist_ok=True)

            input_file = 'input_mol.xyz'
            with open(out_dir/input_file, 'w') as f:
                f.write(xyz_block)

            if solvent is not None:
                subprocess.check_call(
                    ['xtb', input_file, '--opt', '--alpb', solvent, '--parallel', str(num_cores), '--chrg', str(charge)],
                    cwd = out_dir,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.STDOUT,
                )
            else:
                subprocess.check_call(
                    ['xtb', input_file, '--opt', '--parallel', str(num_cores), '--chrg', str(charge)],
                    cwd = out_dir,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.STDOUT,
                )

            opt_xyz_path = out_dir/'xtbopt.xyz'
            if opt_xyz_path.is_file():
                with open(out_dir/'xtbopt.xyz', 'r') as file:
                    xtb_xyz_block = file.readlines()
                    for line in xtb_xyz_block:
                        if 'energy' in line:
                            numbers = re.findall(r"[-+]?(?:\d*\.\d+|\d+)", line)
                            energy = float(numbers[0])
                            break

                if 'energy' in xtb_xyz_block[1]:
                    xtb_xyz_block[1] = '\n' # Replace energy line with blank
                xtb_xyz_block = ''.join(xtb_xyz_block)
            else:
                xtb_xyz_block = None
                energy = None

            with open(out_dir/'charges', 'r') as file:
                lines = file.readlines()
                charges = [0]*len(lines)
                for i, line in enumerate(lines):
                    charges[i] = float(line.split()[0])

        finally:
            if out_dir.exists():
                shutil.rmtree(out_dir)

        return (xtb_xyz_block, energy, charges)


def charges_from_single_point_conformer_with_xtb(conformer: Chem.Mol,
                                                 solvent: Optional[str] = None,
                                                 num_cores: int = 1,
                                                 charge: int = 0,
                                                 temp_dir: str | Path = TMPDIR
                                                 ):
    """
    Compute atomic partial charges from a single point xTB calculation of a provided conformer.

    Uses external calls to GFN2-XTB (command line).

    Parameters
    ----------
    conformer : Chem.Mol
        RDKit mol object containing 3D coordinates.
    solvent : str, optional
        Implicit solvent to be used during calculation. Must be a solvent supported
        by XTB (https://xtb-docs.readthedocs.io/en/latest/gbsa.html). Default is ``None``.
    num_cores : int, optional
        Number of CPU cores to be used in the xTB calculation. Default is 1.
    charge : int, optional
        Molecular charge. Default is 0.
    temp_dir : str or Path, optional
        Temporary directory for I/O. Default is the system temporary directory.

    Returns
    -------
    list
        List of partial charges for each atom (in e-).
    """

    mol = deepcopy(conformer)

    with set_thread_limits(num_cores):
        # rand = str((os.getpid() * int(time.time())) % 123456789)
        rand = str(uuid.uuid4()) + ''.join(str(time.time()).split('.')[1])
        out_dir = Path(f'temp_xtb_opt_{rand}/')
        try:
            if temp_dir != '':
                if isinstance(temp_dir, str):
                    temp_dir = Path(temp_dir)
                out_dir = temp_dir / out_dir
            out_dir.mkdir(exist_ok=True)

            input_file = 'input_mol.xyz'
            Chem.rdmolfiles.MolToXYZFile(mol, str(out_dir/input_file))

            if solvent is not None:
                subprocess.check_call(
                    ['xtb', input_file, '--scc', '--alpb', solvent, '--parallel', str(num_cores), '--chrg', str(charge)],
                    cwd = out_dir,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.STDOUT,
                )
            else:
                subprocess.check_call(
                    ['xtb', input_file, '--scc', '--parallel', str(num_cores), '--chrg', str(charge)],
                    cwd = out_dir,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.STDOUT,
                )

            with open(out_dir/'charges', 'r') as file:
                lines = file.readlines()
                charges = [0]*len(lines)
                for i, line in enumerate(lines):
                    charges[i] = float(line.split()[0])

        finally:
            if out_dir.exists():
                shutil.rmtree(out_dir)

        return charges


def single_point_xtb_from_xyz(xyz_block: str,
                              solvent: Optional[str] = None,
                              num_cores: int = 1,
                              charge: int = 0,
                              temp_dir: str | Path = TMPDIR):
    """
    Compute energy and atomic partial charges from a single point xTB calculation.

    Uses external calls to GFN2-XTB (command line).

    Parameters
    ----------
    xyz_block : str
        String of xyz block representing a molecule.
    solvent : str, optional
        Implicit solvent to be used during calculation. Must be a solvent supported
        by XTB (https://xtb-docs.readthedocs.io/en/latest/gbsa.html). Default is ``None``.
    num_cores : int, optional
        Number of CPU cores to be used in the xTB calculation. Default is 1.
    charge : int, optional
        Molecular charge. Default is 0.
    temp_dir : str or Path, optional
        Temporary directory for I/O. Default is the system temporary directory.

    Returns
    -------
    energy : float
        xTB energy in Hartrees.
    charges : list
        List of partial charges for each atom (in e-).
    """

    with set_thread_limits(num_cores):
        # rand = str((os.getpid() * int(time.time())) % 123456789)
        rand = str(uuid.uuid4()) + ''.join(str(time.time()).split('.')[1])
        out_dir = Path(f'temp_xtb_opt_{rand}/')
        try:
            if temp_dir != '':
                if isinstance(temp_dir, str):
                    temp_dir = Path(temp_dir)
                out_dir = temp_dir / out_dir
            out_dir.mkdir(exist_ok=True)

            input_file = 'input_mol.xyz'
            with open(out_dir/input_file, 'w') as f:
                f.write(xyz_block)

            if solvent is not None:
                output = subprocess.check_output(
                    ['xtb', input_file, '--scc', '--alpb', solvent, '--parallel', str(num_cores), '--chrg', str(charge)],
                    cwd = out_dir,
                    text=True,
                    encoding="utf-8",
                    errors="replace",
                    stderr=subprocess.STDOUT,
                )
            else:
                output = subprocess.check_output(
                    ['xtb', input_file, '--scc', '--parallel', str(num_cores), '--chrg', str(charge)],
                    cwd = out_dir,
                    text=True,
                    encoding="utf-8",
                    errors="replace",
                    stderr=subprocess.STDOUT,
                )

            output = output.split('\n')
            for line in output[::-1]:
                if 'TOTAL ENERGY' in line:
                    numbers = re.findall(r"[-+]?(?:\d*\.\d+|\d+)", line)
                    energy = float(numbers[0])
                    break

            with open(out_dir/'charges', 'r') as file:
                lines = file.readlines()
                charges = [0]*len(lines)
                for i, line in enumerate(lines):
                    charges[i] = float(line.split()[0])
        finally:
            if out_dir.exists():
                shutil.rmtree(out_dir)

        return energy, charges


def _optimize_conformer_worker(params):
    """Worker function for multiprocessing - must be at module level for pickling"""
    return optimize_conformer_with_xtb(*params)

def optimize_conformer_ensemble_with_xtb(conformers: List[Chem.Mol],
                                       solvent: Optional[str] = None,
                                       num_processes: int = 1,
                                       num_workers: int = 1,
                                       charge: int = 0,
                                       temp_dir: str | Path = TMPDIR,
                                       verbose: bool = False):
    """
    GFN2-XTB geometry optimization for a list of conformers.

    Parameters
    ----------
    conformers : list
        List of RDKit Mol objects (with 3D coordinates) to be optimized.
    solvent : str, optional
        Implicit solvent supported by XTB
        (https://xtb-docs.readthedocs.io/en/latest/gbsa.html). Default is ``None``.
    num_processes : int, optional
        Number of CPU cores used per xTB optimization. Default is 1.
    num_workers : int, optional
        Number of parallel workers (processes) to distribute conformers across.
        Ensure num_workers * num_processes <= available CPUs to avoid oversubscription.
        Default is 1.
    charge : int, optional
        Molecular charge. RDKit will be used to compute the formal charge if
        len(conformers) > 1. Default is 0.
    temp_dir : str or Path, optional
        Temporary directory for I/O. Default is the system temporary directory.
    verbose : bool, optional
        Show a simple progress bar in single-process mode. Default is ``False``.

    Returns
    -------
    tuple
        (conformers_opt, energies_opt, charges_opt) - tuple of lists containing
        optimized conformers, their energies, and partial charges.
    """

    if not conformers:
        return [], [], []

    available_cpus = os.cpu_count() or 1
    total_requested_cpus = max(1, num_workers) * max(1, num_processes)
    if total_requested_cpus > available_cpus:
        raise ValueError(
            f"Requested num_workers * num_processes = {total_requested_cpus} exceeds available CPUs ({available_cpus})."
        )

    # Serial path (single conformer or single worker)
    if len(conformers) == 1 or num_workers <= 1:
        conformers_opt = []
        energies_opt = []
        charges_opt = []

        iterator = tqdm(conformers, total=len(conformers), desc='XTB opt') if verbose else conformers
        for conf in iterator:
            opt_conf, opt_energy, opt_charges = optimize_conformer_with_xtb(
                conformer=conf,
                solvent=solvent,
                num_cores=num_processes,
                charge=charge,
                temp_dir=temp_dir,
            )
            conformers_opt.append(opt_conf)
            energies_opt.append(opt_energy)
            charges_opt.append(opt_charges)

        return conformers_opt, energies_opt, charges_opt

    # Parallel
    ctx = multiprocessing.get_context('spawn')

    args = [
        (conf, solvent, num_processes, Chem.GetFormalCharge(conf), temp_dir) for conf in conformers
    ]

    with ctx.Pool(processes=num_workers) as pool:
        iterator = pool.imap(_optimize_conformer_worker, args)
        if verbose:
            iterator = tqdm(iterator, total=len(args), desc='XTB opt')
        results = list(iterator)

    conformers_opt, energies_opt, charges_opt = zip(*results)
    return list(conformers_opt), list(energies_opt), list(charges_opt)


def generate_opt_conformers_xtb(smiles: str,
                                charge: int = 0,
                                solvent: Optional[str]=None,
                                MMFF_optimize: bool = True,
                                num_processes: int = 1,
                                num_workers: int = 1,
                                temp_dir: str | Path = TMPDIR,
                                verbose: bool = False,
                                num_confs: int = 1000):
    """
    Generate conformer ensemble with RDKit then relax with xTB.

    Parameters
    ----------
    smiles : str
        SMILES string of the molecule.
    charge : int, optional
        Molecular charge. Default is 0.
    solvent : str, optional
        Implicit solvent to be used during optimization. Must be a solvent supported
        by XTB (https://xtb-docs.readthedocs.io/en/latest/gbsa.html). Default is ``None``.
    MMFF_optimize : bool, optional
        Optimize RDKit embedded molecules with MMFF94. Default is ``True``.
    num_processes : int, optional
        Number of CPU cores to be used in the xTB geometry optimization. Default is 1.
    num_workers : int, optional
        Number of parallel workers (processes) to distribute conformers across.
        Ensure num_workers * num_processes <= available CPUs to avoid oversubscription.
        Default is 1.
    temp_dir : str or Path, optional
        Temporary directory for I/O. Default is the system temporary directory.
    verbose : bool, optional
        Toggle tqdm progress bar. Default is ``False``.
    num_confs : int, optional
        Number of conformers to initially generate. Default is 1000.

    Returns
    -------
    clustered_conformers_xtb : list
        List of rdkit conformers after xTB relaxation and clustering.
    clustered_energies_xtb : list
        List of energies for associated conformers.
    clustered_charges_xtb : list
        List of partial charges for associated conformers.
    """
    available_cpus = os.cpu_count() or 1
    total_requested_cpus = max(1, num_workers) * max(1, num_processes)
    if total_requested_cpus > available_cpus:
        raise ValueError(
            f"Requested num_workers * num_processes = {total_requested_cpus} exceeds available CPUs ({available_cpus})."
        )
    mol_3d = embed_conformer_from_smiles(smiles, attempts = 50, MMFF_optimize = MMFF_optimize)

    if mol_3d is None:
        return None, None, None

    conformer_ensemble = generate_conformer_ensemble(
        mol_3d,
        num_confs=num_confs,
        num_threads = 4,
        threshold = 0.05,
        num_opt_steps = 50,
    )
    if verbose:
        print(f'Num conformers generated: \t{len(conformer_ensemble)}')

    clustered_conformers_index = cluster_conformers_butina(
        conformer_ensemble,
        threshold = 0.1,
        num_max_conformers = None,
    )
    clustered_conformers = [conformer_ensemble[c] for c in clustered_conformers_index]
    if verbose:
        print(f'Num conformers remaining: \t{len(clustered_conformers)}')

    # further optimizing each conformer with GFN2-XTB
    optimized_conformers_xtb, optimized_energies_xtb, optimized_charges_xtb = optimize_conformer_ensemble_with_xtb(
        clustered_conformers,
        solvent = solvent,
        num_processes = num_processes,
        num_workers = num_workers,
        charge=charge,
        temp_dir = temp_dir,
        verbose=verbose
    )
    if verbose:
        print(f'Num optimized confomers with xtb: \t{len(optimized_conformers_xtb)}')

    clustered_conformers_xtb_index = cluster_conformers_butina(
        optimized_conformers_xtb,
        threshold = 0.1,
        num_max_conformers = None,
    )
    clustered_conformers_xtb = [optimized_conformers_xtb[c] for c in clustered_conformers_xtb_index]
    clustered_charges_xtb = [optimized_charges_xtb[c] for c in clustered_conformers_xtb_index]
    clustered_energies_xtb = [optimized_energies_xtb[c] for c in clustered_conformers_xtb_index]

    if verbose:
        print(f'Num conformers remaining after xtb opt and cluster: \t{len(clustered_conformers_xtb)}')

    return clustered_conformers_xtb, clustered_energies_xtb, clustered_charges_xtb


def generate_opt_conformers(smiles: str,
                            MMFF_optimize: bool = True,
                            verbose: bool = False,
                            num_confs: int = 1000):
    """
    Generate optimal conformers with RDKit (MMFF94).

    Parameters
    ----------
    smiles : str
        SMILES string of the molecule.
    MMFF_optimize : bool, optional
        Optimize RDKit embedded molecules with MMFF94. Default is ``True``.
    verbose : bool, optional
        Toggle tqdm progress bar. Default is ``False``.
    num_confs : int, optional
        Number of conformers to initially generate. Default is 1000.

    Returns
    -------
    list
        List of clustered rdkit conformers after RDKit embedding and optional MMFF relaxation.
    """
    mol_3d = embed_conformer_from_smiles(smiles, attempts = 50, MMFF_optimize = MMFF_optimize)

    if mol_3d is None:
        return None, None, None

    conformer_ensemble = generate_conformer_ensemble(
        mol_3d,
        num_confs=num_confs,
        num_threads = 4,
        threshold = 0.05,
        num_opt_steps = 50,
    )
    if verbose:
        print(f'Num conformers generated: \t{len(conformer_ensemble)}')

    clustered_conformers_index = cluster_conformers_butina(
        conformer_ensemble,
        threshold = 0.1,
        num_max_conformers = None,
    )
    clustered_conformers = [conformer_ensemble[c] for c in clustered_conformers_index]
    if verbose:
        print(f'Num conformers remaining: \t{len(clustered_conformers)}')

    return clustered_conformers



"""
Gaussian volume overlap scoring functions -- Shape-only (i.e., not color)
NUMPY VERSION

Single instance functionality only.

Reference math:
https://doi.org/10.1002/(SICI)1096-987X(19961115)17:14<1653::AID-JCC7>3.0.CO;2-K
https://doi.org/10.1021/j100011a016
"""
import numpy as np
from scipy.spatial import distance

###################################################################################################
####### NUMPY NUMPY NUMPY NUMPY NUMPY NUMPY NUMPY NUMPY NUMPY NUMPY NUMPY NUMPY NUMPY NUMPY #######
###################################################################################################

def VAB_2nd_order_np(centers_1, centers_2, alpha) -> np.ndarray:
    """ 2nd order volume overlap of AB """
    R2 = (distance.cdist(centers_1, centers_2)**2.0).T

    VAB_2nd_order = np.sum(np.pi**(1.5) * np.exp(-(alpha / 2) * R2) / ((2*alpha)**(1.5)))
    return VAB_2nd_order


def shape_tanimoto_np(centers_1, centers_2, alpha) -> np.ndarray:
    """ Compute Tanimoto shape similarity """
    VAA = VAB_2nd_order_np(centers_1, centers_1, alpha)
    VBB = VAB_2nd_order_np(centers_2, centers_2, alpha)
    VAB = VAB_2nd_order_np(centers_1, centers_2, alpha)
    return VAB / (VAA + VBB - VAB)


def get_overlap_np(centers_1:np.ndarray,
                   centers_2:np.ndarray,
                   alpha:float = 0.81
                   ) -> np.ndarray:
    """ NumPy implementation of shape similarity via gaussian overlaps (single instance) """
    tanimoto = shape_tanimoto_np(centers_1, centers_2, alpha)
    return tanimoto


def VAB_2nd_order_cosine_np(centers_1: np.ndarray,
                            centers_2: np.ndarray,
                            vectors_1: np.ndarray,
                            vectors_2: np.ndarray,
                            alpha: float,
                            allow_antiparallel: bool,
                            ) -> np.ndarray:
    """
    2nd order volume overlap of AB weighted by cosine similarity.
    NumPy implementation with single instance functionality.
    """
    if len(centers_1.shape) == 2:
        # Normalize vectors for cosine similarity
        norm_v1 = np.linalg.norm(vectors_1, axis=1, keepdims=True)
        norm_v2 = np.linalg.norm(vectors_2, axis=1, keepdims=True)

        # Avoid division by zero if a vector is all zeros
        vec1_norm = np.divide(vectors_1, norm_v1, out=np.zeros_like(vectors_1), where=norm_v1!=0)
        vec2_norm = np.divide(vectors_2, norm_v2, out=np.zeros_like(vectors_2), where=norm_v2!=0)

        # cosine similarity
        V2 = np.matmul(vec1_norm, vec2_norm.T).T # Now uses normalized vectors
        if allow_antiparallel:
            V2 = np.abs(V2)
        else:
            V2 = np.clip(V2, 0., 1.)

        V2 = (V2 + 2.)/3. # Following PheSA's suggestion for weighting

        R2 = (distance.cdist(centers_1, centers_2)**2.0).T

        VAB_second_order = np.sum(np.pi**(1.5) * V2 * np.exp(-(alpha / 2) * R2) / ((2*alpha)**(1.5)))

    return VAB_second_order



"""
Protonation functions.

Utilizing `molscrub_utils requires `molscrub` package.
Instructions for installation can be found at [https://github.com/forlilab/molscrub](https://github.com/forlilab/molscrub).

Utilizing `chemaxon_utils` requires ChemAxon license.
"""
